import os
import traceback
from functools import partial
from pkg_resources import resource_filename
import argparse

from PyQt5.QtWidgets import QMessageBox, QStyle, QTableWidget, \
    QTableWidgetItem, QMainWindow, QDialog, QDialogButtonBox, \
    QListWidgetItem, QMenu, QAction
from PyQt5 import QtGui
from PyQt5.QtCore import Qt, QThread, pyqtSignal, pyqtSlot
from PyQt5 import uic

import numpy as np
# import scipy.signal
import sklearn.cluster
import matplotlib
matplotlib.use('QT5Agg')
import matplotlib.pyplot as plt

from .decomp.decompworker import DecompWorker
# import octavvs.io
from octavvs.algorithms import normalization
from octavvs.io import DecompositionData, Parameters
from octavvs.ui import (FileLoader, ImageVisualizer, OctavvsMainWindow,
                        NoRepeatStyle)



DecompositionMainWindow = uic.loadUiType(resource_filename(
    __name__, "decomp/decomposition.ui"))[0]
SettingsTableWindow = uic.loadUiType(resource_filename(
    __name__, "decomp/settings_table.ui"))[0]

class DialogSettingsTable(QDialog, SettingsTableWindow):
    def __init__(self, parent=None):
        QDialog.__init__(self, parent)
        self.setupUi(self)

class MyMainWindow(ImageVisualizer, FileLoader, OctavvsMainWindow,
                   DecompositionMainWindow):

    closing = pyqtSignal()
    startDecomp = pyqtSignal(DecompositionData, Parameters)
    startBatch = pyqtSignal(DecompositionData, Parameters)

    @classmethod
    def program_name(cls):
        "Return the name of the program that this main window represents"
        return 'Decomposition'

    def __init__(self, parent=None, files=None, paramFile=None):
        super().__init__(parent)

        self.data = DecompositionData()
        self.workerRunning = False

        # Avoid repeating spinboxes
        self.spinBoxComponents.setStyle(NoRepeatStyle())
        self.spinBoxIterations.setStyle(NoRepeatStyle())

        # Not really needed
        self.pushButtonSaveParameters.clicked.connect(self.saveParameters)
        self.pushButtonLoadParameters.clicked.connect(self.loadParameters)

        self.imageVisualizer.comboBoxCmaps.currentTextChanged.connect(
            self.plot_roi.set_cmap)
        self.imageVisualizer.plot_raw.updatedProjection.connect(
            self.plot_roi.set_data)
        self.pushButtonRoiClear.clicked.connect(self.roiClear)
        self.pushButtonRoiDraw.clicked.connect(self.roiDrawFree)
        self.pushButtonRoiAdd.clicked.connect(self.roiAddArea)
        self.pushButtonRoiRemove.clicked.connect(self.roiRemoveArea)
        self.pushButtonRoiErase.clicked.connect(
            self.plot_roi.erase_last_point)
        self.pushButtonRoiInvert.clicked.connect(self.plot_roi.invert_roi)
        self.plot_roi.updated.connect(self.roiUpdateSelected)

        # Connect all the data loading/saving stuff
        for att, what in {'Roi': 'roi', 'Dc': 'decomposition',
                          'Ca': 'clustering'}.items():
            getattr(self, 'pushButton%sDirectory' % att).clicked.connect(
                partial(self.genericDirSelect, what=what))
            getattr(self, 'lineEdit%sDirectory' % att).editingFinished.connect(
                partial(self.reloadCheck, what=what))
            getattr(self, 'pushButton%sLoad' % att).clicked.connect(
                partial(self.genericLoad, what=what))
            getattr(self, 'pushButton%sSave' % att).clicked.connect(
                partial(self.genericSave, what=what))

        self.imageVisualizer.plot_spectra.updated.connect(self.updateDC)
        self.pushButtonStart.clicked.connect(self.startDC)
        self.pushButtonStop.clicked.connect(self.stopDC)
        self.pushButtonRun.clicked.connect(self.dcBatchRun)
        self.pushButtonRunStop.clicked.connect(self.stopDC)

        self.pushButtonSettingsInfo.clicked.connect(self.dcSettingsShow)
        self.dialogSettingsTable = DialogSettingsTable()
        self.dialogSettingsTable.buttonBox.clicked.connect(
            self.dcSettingsClicked)
        # todo: Preset editing - save in application settings?

        self.comboBoxInitialValues.currentIndexChanged.connect(
            self.dcInitialValuesChanged)
        self.lineEditInitSkew.setFormat('%g')
        self.lineEditInitSkew.setRange(1, 1000)
        self.lineEditInitNoise.setFormat('%g')
        self.lineEditInitNoise.setRange(.001, 1)
        self.comboBoxDerivative.currentIndexChanged.connect(
            self.dcDerivativeChanged)
        self.checkBoxContrast.toggled.connect(self.dcContrastChanged)
        self.lineEditContrast.setFormat("%g")
        self.lineEditContrast.setRange(0, .1)
        self.lineEditContrast.editingFinished.connect(self.dcContrastEdit)
        self.horizontalSliderContrast.valueChanged.connect(
            self.dcContrastSlide)

        self.imageVisualizer.comboBoxCmaps.currentTextChanged.connect(
            self.plot_decomp.set_cmap)
        self.comboBoxPlotMode.currentIndexChanged.connect(
            self.plot_decomp.set_display_mode)
        self.plot_decomp.displayModesUpdated.connect(
            self.updateDCPlotModes)

        self.pushButtonCluster.clicked.connect(self.caStartClustering)
        # self.toolButtonAnnotationPlus.setIcon(
        #     self.style().standardIcon(QStyle.SP_DialogYesButton))
        self.toolButtonAnnotationMinus.setIcon(
            self.style().standardIcon(QStyle.SP_TrashIcon))
        self.toolButtonAnnotationPlus.clicked.connect(self.caAddAnnotation)
        self.toolButtonAnnotationMinus.clicked.connect(self.caDelAnnotation)
        self.listWidgetClusterAnnotations.itemSelectionChanged.connect(
            self.caSelectAnnotation)
        self.listWidgetClusterAnnotations.itemChanged.connect(
            self.caEditedAnnotation)
        self.listWidgetClusterUsed.itemDoubleClicked.connect(
            self.caClickedUsed)
        self.listWidgetClusterUnused.itemDoubleClicked.connect(
            self.caClickedUnused)
        self.plot_cluster.clicked.connect(self.caSelectCluster)

        exportMenu = QMenu()
        for txt, a, b in [['Annotation averages', 'annotation', False],
                         ['Cluster averages', 'cluster', False],
                         ['Individual spectra', None, False],
                         ['Batch, annotation averages', 'annotation', True],
                         ['Batch, cluster averages', 'cluster', True],
                         ['Batch, individual spectra', None, True]]:
            exportAction = QAction(txt, self)
            exportAction.triggered.connect(
                partial(self.caExport, average=a, batch=b))
            exportAction.setIconVisibleInMenu(False)
            exportMenu.addAction(exportAction)
        self.toolButtonClusterExport.setMenu(exportMenu)

        self.worker = DecompWorker()
        self.workerThread = QThread()
        self.worker.moveToThread(self.workerThread)
        self.closing.connect(self.workerThread.quit)
        self.startDecomp.connect(self.worker.decompose)
        self.worker.done.connect(self.dcDone)
        self.worker.stopped.connect(self.dcStopped)
        self.worker.failed.connect(self.dcFailed)
        self.worker.progress.connect(self.dcProgress)
        self.worker.progressPlot.connect(self.dcProgressPlot)
        self.worker.batchProgress.connect(self.dcBatchProgress)
        self.worker.fileLoaded.connect(self.updateFile)
        self.worker.loadFailed.connect(self.showLoadErrorMessage)
        self.startBatch.connect(self.worker.startBatch)
        self.worker.batchDone.connect(self.dcBatchDone)
        self.workerThread.start()

        self.lineEditTolerance.setFormat("%g")
        self.lineEditTolerance.setRange(1e-10, 1)
        # self.lineEditRelError.setFormat("%.4g")

        self.updateWavenumberRange()

        self.post_setup()
        if paramFile is not None: # Loads the parameter file passed as argument
            self.loadParameters(filename=paramFile)
        if files is not None and files != []:
            self.updateFileList(files, False) # Load files passed as arguments


    def closeEvent(self, event):
        self.dialogSettingsTable.close()
        self.autosaveCheck()
        self.worker.halt = True
        self.closing.emit()
        plt.close('all')
        self.workerThread.wait()
        self.deleteLater()


    def loadFolder(self, *argv, **kwargs):
        assert self.workerRunning == 0
        super().loadFolder(*argv, **kwargs)

    def updateFileList(self, *argv, **kwargs):
        ret = super().updateFileList(*argv, **kwargs)
        if ret:
            dups = self.data.get_duplicate_filenames()
            if dups:
                q = QMessageBox(self)
                q.setIcon(QMessageBox.Warning)
                q.setWindowTitle('Warning: identical filenames')
                q.setText('Some input files have identical names, which '
                          'will cause problems with overwritten output files.')
                q.setTextFormat(Qt.PlainText)
                q.setDetailedText('Examples:\n' + '\n'.join(list(dups)[:10]))
                q.exec()
        return ret

    def loadFile(self, file):
        "Save ROI before proceeding"
        self.autosaveCheck()
        return super().loadFile(file)

    def populatePlots(self):
        "Send data to plots etc after loading rdc"
        self.plot_roi.set_roi(self.data.roi)
        self.dcSettingsUpdate()
        self.plot_decomp.clear_and_set_roi(self.data.decomposition_roi)
        self.plot_decomp.set_errors(self.data.decomposition_errors)
        dd = self.data.decomposition_data
        if dd:
            if 0 in dd:
                self.plot_decomp.set_initial_spectra(dd[0]['spectra'])
            lastiter = max(dd.keys())
            ds = dd[lastiter]
            self.plot_decomp.set_concentrations(ds['concentrations'])
            self.plot_decomp.set_spectra(ds['spectra'])
        self.plot_cluster.set_roi_and_clusters(
            self.data.clustering_roi, self.data.clustering_labels)
        self.caPopulateLists()

    @pyqtSlot(int)
    def updateFile(self, num):
        super().updateFile(num)
        self.updateWavenumberRange()
        self.plot_roi.set_basic_data(
            wh=self.data.wh, pixelxy=self.data.pixelxy)
        self.plot_decomp.set_basic_data(
            wn=self.imageVisualizer.plot_raw.getWavenumbers(),
            wh=self.data.wh, pixelxy=self.data.pixelxy)
        self.plot_cluster.set_basic_data(
            wh=self.data.wh, pixelxy=self.data.pixelxy)
        self.autoLoad()
        self.populatePlots()
        self.updatedFile()

    @pyqtSlot(str, str, str)
    def showLoadErrorMessage(self, file, err, details):
        """
        Error message from loading a file in the worker thread,
        with abort option
        """
        warning=False
        q = self.loadErrorBox(file, (err, details if details else None),
                              warning)
        q.addButton('Ignore' if warning else 'Skip file',
                    QMessageBox.AcceptRole)
        abort = q.addButton('Abort', QMessageBox.AcceptRole)
        q.exec()
        if q.clickedButton() == abort:
            self.stopDC()

    def updateDimensions(self, wh):
        super().updateDimensions(wh)
        self.plot_roi.adjust_geometry(wh)
        self.plot_decomp.adjust_geometry(wh)

    def setPlotColors(self, cmap):
        super().setPlotColors(cmap)
        self.plot_decomp.draw_idle()

    # Loading and saving data
    def suggestedFilename(self, what):
        if not self.data.curFile:
            return None
        fn = os.path.splitext(os.path.basename(self.data.curFile))[0]
        if what == 'roi':
            return os.path.join(self.lineEditRoiDirectory.text(), fn) + '.roi'
        if what == 'decomposition':
            return os.path.join(self.lineEditDcDirectory.text(), fn) + '.dcn'
        if what == 'clustering':
            return os.path.join(self.lineEditCaDirectory.text(), fn) + '.can'

    def fileFormats(self, what):
        if what == 'roi':
            ff = "ROI HDF5 files (*.roi);;ROI MATLAB files (*.mat);;"\
                "ROI CVS files (*.csv);;"
            descr = 'ROI'
        elif what == 'decomposition':
            ff = "Decomposition HDF5 files (*.dcn);;"
            descr = 'decomposition data'
        elif what == 'clustering':
            ff = "Clustering/annotation HDF5 files (*.can);;"
            descr = 'clustering/annotations'
        else:
            ff = ""
            descr = 'data'
        return ff + "Combined HDF5 files (*.odd);;All files (*)", descr

    def genericLoad(self, what='all'):
        if not self.data.curFile:
            return
        path = os.path.split(self.suggestedFilename(what))
        filter, description = self.fileFormats(what)
        filename = self.getLoadFileName(
            "Load %s from file" % description,
            filter=filter,
            directory=path[0], defaultfilename=path[1])
        if not filename:
            return
        self.data.load_rdc(filename, what=what)
        self.populatePlots()

    def autoLoad(self):
        for what in ['roi', 'decomposition', 'clustering']:
            filename = self.suggestedFilename(what)
            if filename and os.path.exists(filename):
                self.data.load_rdc(filename, what=what)

    def genericSave(self, what='all'):
        if not self.data.curFile:
            return
        path = os.path.split(self.suggestedFilename(what))
        filter, description = self.fileFormats(what)
        filename = self.getSaveFileName(
            "Save %s to file" % description,
            filter=filter,
            directory=path[0], defaultfilename=path[1])
        if not filename:
            return

        # Make sure self.data is up-to-date
        if what in ['roi', 'all']:
            self.data.set_roi(self.plot_roi.get_roi())
        if what in ['decomposition', 'all']:
            ...
        if what in ['clustering', 'all']:
            self.caSelectAnnotation()
        ext = os.path.splitext(filename)[1]
        if what == 'roi':
            self.data.save_roi(filename=filename, fmt=ext)
        else:
            self.data.save_rdc(filename=filename, what=what)

    def autosaveCheck(self):
        if self.data.curFile and not self.workerRunning:
            save = []
            if self.checkBoxRoiAutosave.isChecked():
                self.data.set_roi(self.plot_roi.get_roi())
                save.append('roi')
            if self.checkBoxDcAutosave.isChecked():
                save.append('decomposition')
            if self.checkBoxCaAutosave.isChecked():
                self.caSelectAnnotation()
                save.append('clustering')
            for what in save:
                self.data.save_rdc(self.suggestedFilename(
                    what=what), what=what)

    def reloadCheck(self, what):
        filename = self.suggestedFilename(what)
        if filename:
            descr = self.fileFormats(what)[1]
            if os.path.exists(filename):
                yn = QMessageBox.question(
                    self, 'Directory changed',
                    '(Re)load saved %s from selected directory?' % descr)
                if yn == QMessageBox.Yes:
                    self.data.load_rdc(filename, what=what)
                    self.populatePlots()

    def genericDirSelect(self, what):
        descr = self.fileFormats(what)[1]
        short = {'roi': 'roi', 'decomposition': 'dc',
                 'clustering': 'ca', 'all': 'rdc'}[what]
        rdir = self.getDirectoryName(
            'Select %s directory' % descr, settingname=short+'Dir',
            default=os.path.dirname(self.data.curFile))
        if rdir is not None:
            Short = short[0].upper() + short[1:]
            getattr(self, 'lineEdit%sDirectory' % Short).setText(rdir)
            self.reloadCheck(what=what)

    # Roi, Region of interest
    def roiClear(self):
        self.roiDrawFree(True)
        self.plot_roi.clear()

    def roiCheckButtons(self):
        mode = self.plot_roi.get_draw_mode()
        self.pushButtonRoiDraw.setChecked(mode == 'click')
        self.pushButtonRoiAdd.setChecked(mode == 'add')
        self.pushButtonRoiRemove.setChecked(mode == 'remove')

    def roiDrawFree(self, checked):
        self.plot_roi.set_draw_mode('click')
        self.roiCheckButtons()

    def roiAddArea(self, checked):
        self.plot_roi.set_draw_mode('add' if checked else 'click')
        self.roiCheckButtons()

    def roiRemoveArea(self, checked):
        self.plot_roi.set_draw_mode('remove' if checked else 'click')
        self.roiCheckButtons()

    def roiUpdateSelected(self, n, m):
        self.labelRoiSelected.setText('Selected: %d / %d' % (n, m))


    # DC, Decomposition
    def updateDC(self):
        "Respond to changes by user in UI"
        if self.workerRunning:
            return
        self.pushButtonStart.setEnabled(True)

    @pyqtSlot(list)
    def updateDCPlotModes(self, modes):
        "Refresh the list of plot modes"
        # Quick and dirty
        ct = self.comboBoxPlotMode.currentText()
        self.comboBoxPlotMode.clear()
        self.comboBoxPlotMode.insertItems(0, modes)
        try:
            self.comboBoxPlotMode.setCurrentText(ct)
        except ValueError:
            pass

    def toggleRunning(self, newstate):
        "Move between states: 0=idle, 1=run one, 2=run batch"
        onoff = not newstate
        self.fileLoader.setEnabled(onoff)
        self.pushButtonRoiLoad.setEnabled(onoff)
        self.pushButtonRoiSave.setEnabled(onoff)
        self.pushButtonDcLoad.setEnabled(onoff)
        self.pushButtonDcSave.setEnabled(onoff)
        self.pushButtonCaLoad.setEnabled(onoff)
        self.pushButtonCaSave.setEnabled(onoff)
        self.pushButtonRun.setEnabled(onoff)
        self.pushButtonStart.setEnabled(onoff)
        # if newstate:
        #     self.progressBarIteration.setValue(0)
        self.pushButtonStop.setEnabled(newstate == 1)
        self.pushButtonRunStop.setEnabled(newstate == 2)
        self.workerRunning = newstate

    def getDCSettings(self):
        "Get relevant parameters for decomposition, as a dict"
        params = self.getParameters()
        return params.filtered('dc')

    def startDC(self):
        "Decompose the current data only"
        if self.workerRunning:
            return
        params = Parameters(self.getDCSettings().items())
        self.data.set_roi(self.plot_roi.get_roi())
        if params.dcRoi == 'require' and self.data.roi is None:
            self.errorMsg.showMessage(
                'Non-empty ROI is required')
            return
        self.data.set_decomposition_settings(params)
        self.dcSettingsUpdate()
        self.plot_decomp.clear_and_set_roi(self.data.decomposition_roi)
        # self.progressBarIteration.setMaximum(params.dcIterations)
        # self.progressBarIteration.setFormat('initializing')
        self.lineEditIteration.setText('init')
        self.toggleRunning(1)
        self.startDecomp.emit(self.data, params)

    def clearDC(self):
        if self.workerRunning == 1:
            return
        self.lineEditIteration.setText('idle')
        # self.progressBarIteration.setValue(0)
        # self.progressBarIteration.setFormat('idle')

    def stopDC(self):
        self.worker.halt = True
        self.pushButtonStop.setEnabled(False)
        self.pushButtonRunStop.setEnabled(False)

    @pyqtSlot()
    def dcStopped(self):
        self.worker.halt = False
        if self.workerRunning == 1:
            self.toggleRunning(0)
        self.updateDC()

    @pyqtSlot(str, str)
    def dcFailed(self, err, trace):
        self.errorMsg.showMessage(
            'Processing failed:<pre>\n' + err + "\n\n" + trace + '</pre>')
        self.dcStopped()

    @pyqtSlot(np.ndarray, np.ndarray, np.ndarray)
    def dcDone(self, spectra, concentrations, errors):
        self.plot_decomp.set_spectra(spectra)
        self.plot_decomp.set_concentrations(concentrations)
        self.plot_decomp.set_errors(errors)
        self.lineEditIteration.setText('%d/done' % len(errors))
        self.dcStopped()
        self.caClear()

    @pyqtSlot(int, float)
    def dcProgress(self, iteration, error):
        if iteration:
            self.lineEditIteration.setText('%d' % iteration)
            self.lineEditRelError.setText('%.4f' % (error * 100))

    @pyqtSlot(int, np.ndarray, np.ndarray, list)
    def dcProgressPlot(self, iteration, spectra, concentrations, errors):
        if iteration:
            self.plot_decomp.set_concentrations(concentrations)
            self.plot_decomp.set_spectra(spectra)
        else:
            # self.data.add_decomposition_data(0, spectra, None)
            self.plot_decomp.set_initial_spectra(spectra)
        self.plot_decomp.set_errors(errors)

    def dcInitialValuesChanged(self):
        self.stackedWidgetInit.setCurrentIndex(
            self.comboBoxInitialValues.currentIndex())

    def dcDerivativeChanged(self):
        en = self.comboBoxDerivative.currentIndex() > 0
        self.spinBoxDerivativePoly.setEnabled(en)
        self.spinBoxDerivativeWindow.setEnabled(en)

    def dcContrastChanged(self):
        en = self.checkBoxContrast.isChecked()
        self.comboBoxContrast.setEnabled(en)
        self.horizontalSliderContrast.setEnabled(en)
        self.lineEditContrast.setEnabled(en)

    def dcContrastSlide(self):
        self.lineEditContrast.setValue(
            self.horizontalSliderContrast.value() * .001)

    def dcContrastEdit(self):
        self.horizontalSliderContrast.setValue(
            self.lineEditContrast.value() * 1000)

    def dcSettingsUpdate(self):
        "Update the table of most recent dc settings"
        dds = self.data.decomposition_settings
        if dds is None:
            dds = {}
        tw = self.dialogSettingsTable.tableWidget
        tw.setRowCount(len(dds))
        for i, k in enumerate(dds.keys()):
            tw.setItem(i, 0, QTableWidgetItem(k))
            tw.setItem(i, 1, QTableWidgetItem(str(dds[k])))
        tw.resizeColumnsToContents()

    def dcSettingsShow(self, button):
        "Show the table of dc settings"
        self.dialogSettingsTable.show()
        self.dialogSettingsTable.raise_()

    def dcSettingsClicked(self, button):
        if button == QDialogButtonBox.Apply:
            ...
        elif button == QDialogButtonBox.Close:
            self.dialogSettingsTable.hide()

    def dcBatchRun(self, checked=False):
        """
        Run the selected preprocessing steps, showing a dialog if no output
        path is given. The 'checked' parameter is just a placeholder to
        match QPushButton.clicked().
        """
        assert not self.workerRunning
        if len(self.data.filenames) < 1:
            self.errorMsg.showMessage('Load some data first')
            return

        params = self.getParameters()
        self.toggleRunning(2)
        self.startBatch.emit(self.data, params)

    @pyqtSlot(int, int)
    def dcBatchProgress(self, a, b):
        self.progressBarRun.setValue(a)
        self.progressBarRun.setMaximum(b-1)

    @pyqtSlot(bool)
    def dcBatchDone(self, success):
        self.worker.halt = False
        self.lineEditIteration.setText('done' if success else 'fail')
        self.toggleRunning(0)


    # CA, Clustering and annotation
    def caStartClustering(self):
        if self.data.raw is None:
            return
        p = self.getParameters()
        if p.caInput == 'decomposition':
            dd = self.data.decomposition_data
            if dd is None:
                return
            lastiter = max(dd.keys())
            if not lastiter:
                return # Only initial spectra available
            inp = dd[lastiter]['concentrations'].T
            roi = self.data.decomposition_roi
        else:
            inp = self.data.raw
            roi = self.data.roi if p.caInput == 'raw-roi' else None
            if roi is not None:
                inp = inp[roi]
        inp = normalization.normalize_features(inp, p.caNormalization)

        if p.caMethod == 'kmeans':
            clusterer = sklearn.cluster.KMeans(
                n_clusters=p.caClusters, n_init=p.caRestarts)
            labels = clusterer.fit_predict(inp)
        elif p.caMethod == 'mbkmeans':
            clusterer = sklearn.cluster.MiniBatchKMeans(
                n_clusters=p.caClusters, n_init=p.caRestarts)
            labels = clusterer.fit_predict(inp)
        elif p.caMethod == 'strongest':
            labels = inp.argmax(1)
            p.caClusters = labels.max() + 1
        else:
            raise ValueError('Unknown clustering method %s' % p.caMethod)
        # elif p.caMethod == 'meanshift':
        #     clusterer = sklearn.cluster.MeanShift(n_jobs=-1)
        # elif p.caMethod == 'spectral':
        #     clusterer = sklearn.cluster.SpectralClustering(
        #         n_clusters=p.caClusters, n_init=p.caRestarts, n_jobs=-1)
        # labels = clusterer.fit_predict(inp)
        # if p.caMethod == 'meanshift':
        #     p.caClusters = labels.max() + 1
        self.data.set_clustering_settings(p, roi)
        self.data.set_clustering_labels(labels, relabel=True)
        self.plot_cluster.set_roi_and_clusters(
            self.data.clustering_roi, self.data.clustering_labels)
        self.caPopulateLists()

    def caAddAnnotation(self, checked=False, name='New annotation',
                        edit=True):
        # self.caSelectAnnotation()
        # self.selectedAnnotation = None
        item = QListWidgetItem(name, parent=self.listWidgetClusterAnnotations)
        item.setFlags(item.flags() | Qt.ItemIsEditable)
        item.octavvs_id = name
        self.listWidgetClusterAnnotations.addItem(item)
        if edit:
            self.listWidgetClusterAnnotations.setCurrentItem(item)
            self.listWidgetClusterAnnotations.editItem(item)

    def caClustersToList(self, clusters, listw):
        inlist = set()
        for r in reversed(range(listw.count())):
            c = listw.item(r).octavvs_id
            if c not in clusters:
                listw.takeItem(r)
            else:
                inlist.add(c)
        for c in clusters - inlist:
            item = QListWidgetItem('Cluster %d' % c)
            cc = np.asarray(self.plot_cluster.cluster_color(c)[:3]) * 255
            item.setBackground(QtGui.QBrush(QtGui.QColor(*cc)))
            item.octavvs_id = c
            listw.addItem(item)

    def caClearList(self, listw):
        while listw.count():
            listw.takeItem(0)

    def caPopulateLists(self):
        for l in [self.listWidgetClusterUsed, self.listWidgetClusterUnused ]:
            self.caClearList(l)
        awlist = self.listWidgetClusterAnnotations
        awlist.setCurrentItem(None)
        self.listWidgetClusterUsed.setEnabled(False)
        anns = set()
        for r in reversed(range(awlist.count())):
            name = awlist.item(r).text()
            # Remove duplicates. Should maybe also remove if not in self.data
            if name in anns:
                awlist.takeItem(r)
            else:
                anns.add(name)
        if self.data.clustering_annotations:
            for name in self.data.clustering_annotations:
                if name not in anns:
                    self.caAddAnnotation(name=name, edit=False)
        if self.data.clustering_labels is not None:
            self.caClustersToList(self.data.get_unannotated_clusters(),
                                  self.listWidgetClusterUnused)

    def caUpdateAnnotation(self):
        "Save selected clusters to self.data"
        sel = self.listWidgetClusterAnnotations.selectedItems()
        sel = sel[0].text() if sel else None
        if sel:
            clusters = set()
            for r in range(self.listWidgetClusterUsed.count()):
                clusters.add(self.listWidgetClusterUsed.item(r).octavvs_id)
            self.data.set_annotation_clusters(sel, clusters)
        else:
            assert not self.listWidgetClusterUsed.isEnabled()

    def caSelectAnnotation(self):
        "Update the currently selected annotation"
        sel = self.listWidgetClusterAnnotations.selectedItems()
        sel = sel[0].text() if sel else None
        if sel:
            used = self.data.get_annotation_clusters(sel)
            self.caClustersToList(used, self.listWidgetClusterUsed)
        else:
            self.caClearList(self.listWidgetClusterUsed)
        self.listWidgetClusterUsed.setEnabled(sel is not None)

    def caStoreAnnotations(self):
        "Save all to self.data"
        self.caUpdateAnnotation()
        awlist = self.listWidgetClusterAnnotations
        for r in range(awlist.count()):
            name = awlist.item(r).text()
            self.data.set_annotation_clusters(name)

    def caDelAnnotation(self):
        "Delete annotation and make items unused"
        sel = self.listWidgetClusterAnnotations.selectedItems()
        if not sel:
            return
        self.data.del_annotation(sel[0].text())
        while self.listWidgetClusterUsed.count():
            item = self.listWidgetClusterUsed.takeItem(0)
            self.listWidgetClusterUnused.addItem(item)
        self.listWidgetClusterAnnotations.takeItem(
            self.listWidgetClusterAnnotations.row(sel[0]))

    def caEditedAnnotation(self, item):
        "Enact annotation name change"
        nnew = item.text()
        clusters = self.data.get_annotation_clusters(nnew)
        if hasattr(item, 'octavvs_id'):
            nold = item.octavvs_id
            clusters.update(self.data.get_annotation_clusters(nold))
            self.data.del_annotation(nold)
        item.octavvs_id = nnew
        self.data.set_annotation_clusters(nnew, clusters)
        self.caUpdateAnnotation()

    def caClickedUsed(self, item):
        if not self.listWidgetClusterAnnotations.selectedItems():
            return
        self.listWidgetClusterUsed.takeItem(
            self.listWidgetClusterUsed.row(item))
        self.listWidgetClusterUnused.addItem(item)
        self.caUpdateAnnotation()

    def caClickedUnused(self, item):
        if not self.listWidgetClusterAnnotations.selectedItems():
            return
        self.listWidgetClusterUnused.takeItem(
            self.listWidgetClusterUnused.row(item))
        self.listWidgetClusterUsed.addItem(item)
        self.caUpdateAnnotation()

    def caSelectCluster(self, num):
        if not self.listWidgetClusterAnnotations.selectedItems():
            return
        for r in range(self.listWidgetClusterUnused.count()):
            item = self.listWidgetClusterUnused.item(r)
            if item.octavvs_id == num:
                item = self.listWidgetClusterUnused.takeItem(
                    self.listWidgetClusterUnused.row(item))
                self.listWidgetClusterUsed.addItem(item)
                self.caUpdateAnnotation()
                break

    def caClear(self):
        "A bit brutal as response to changes, maybe"
        input = self.caInputNames[self.comboBoxClusterInput.currentIndex()]
        enabled = self.data.raw is not None and (
            input == 'raw' or self.data.decomposition_data is not None)
        self.pushButtonCluster.setEnabled(enabled)
        self.plot_cluster.clear_clusters()
        for l in [self.listWidgetClusterUsed, self.listWidgetClusterUnused]:
            while l.count():
                l.takeItem(0)

    def caExport(self, average=None, batch=False):
        if not self.data.curFile:
            return
        if batch:
            dirname = self.getDirectoryName(
                "Batch export annotated spectra",
                settingname='exportDir')
            for i in range(len(self.data.filenames)):
                self.fileLoader.spinBoxFileNumber.setValue(i + 1)
                filename = os.path.join(
                    dirname, os.path.splitext(
                        os.path.basename(self.data.curFile))[0]+'.csv')
                try:
                    self.data.save_annotated_spectra(
                        filename, filetype='csv', average=average)
                except Exception as e:
                    self.showDetailedErrorMessage(
                            "Error saving settings to "+filename+": "+repr(e),
                            traceback.format_exc())
                    return
        else:
            filename = os.path.splitext(self.data.curFile)[0]+'.csv'
            filename = self.getSaveFileName(
                    "Export annotated spectra",
                    filter="Comma-separated values (*.csv);;All files (*)",
                    settingname='exportDir',
                    suffix='csv',
                    defaultfilename=filename)
            if not filename:
                return
            try:
                self.data.save_annotated_spectra(
                    filename, filetype='csv', average=average)
            except Exception as e:
                self.showDetailedErrorMessage(
                        "Error saving settings to "+filename+": "+repr(e),
                        traceback.format_exc())


    # Parameter handling
    useRoiNames = ['ignore', 'ifdef', 'require']
    dcInitialValuesNames = ['simplisma', 'kmeans', 'clustersubtract']
    dcAlgorithmNames = ['mcr-als-anderson', 'mcr-als']
    caInputNames = ['raw-roi', 'raw', 'decomposition']
    caMethodNames = ['kmeans', 'mbkmeans', 'strongest']
    caNormalizationNames = ['none', 'mean1', 'mean0', 'mean0var1']
    # caMethodNames = ['kmeans', 'meanshift', 'spectral']

    # Loading and saving parameters
    def getParameters(self):
        """Copy from UI to some kind of parameters object."""
        p = Parameters()
        self.fileLoader.saveParameters(p)
        self.imageVisualizer.saveParameters(p)

        # p.dirMode = self.comboBoxDirectory.currentIndex()
        # p.directory = self.lineEditDirectory.text()
        # p.autosave = self.checkBoxRdcAutosave.isChecked()

        p.dcAlgorithm = self.dcAlgorithmNames[
            self.comboBoxAlgorithm.currentIndex()]
        p.dcImpute = self.checkBoxImpute.isChecked()
        p.dcDerivative = self.comboBoxDerivative.currentIndex()
        p.dcDerivativeWindow = self.spinBoxDerivativeWindow.value()
        p.dcDerivativePoly = self.spinBoxDerivativePoly.value()
        p.dcComponents = self.spinBoxComponents.value()
        p.dcStartingPoint = self.comboBoxStartingPoint.currentIndex()
        p.dcInitialValues = self.dcInitialValuesNames[
            self.comboBoxInitialValues.currentIndex()]
        p.dcInitialValuesSkew = self.lineEditInitSkew.value()
        p.dcInitialValuesPower = self.spinBoxInitPower.value()
        p.dcInitialValuesNoise = self.lineEditInitNoise.value()
        p.dcRoi = self.useRoiNames[self.comboBoxUseRoi.currentIndex()]
        p.dcContrast = self.checkBoxContrast.isChecked()
        p.dcContrastConcentrations = self.comboBoxContrast.currentIndex()
        p.dcContrastWeight = self.lineEditContrast.value()
        p.dcIterations = self.spinBoxIterations.value()
        p.dcTolerance = self.lineEditTolerance.value()

        p.caInput = self.caInputNames[self.comboBoxClusterInput.currentIndex()]
        p.caNormalization = self.caNormalizationNames[
            self.comboBoxClusterNormalization.currentIndex()]
        p.caMethod = self.caMethodNames[
            self.comboBoxClusterMethod.currentIndex()]
        p.caClusters = self.spinBoxClusterClusters.value()
        p.caRestarts = self.spinBoxClusterRestarts.value()
        p.caAnnotations = []
        for i in range(self.listWidgetClusterAnnotations.count()):
            item = self.listWidgetClusterAnnotations.item(i)
            p.caAnnotations.append(item.text())

        return p

    def saveParameters(self):
        filename = self.getSaveFileName(
                "Save decomposition settings",
                filter="Setting files (*.djs);;All files (*)",
                settingname='settingsDir',
                suffix='djs')
        if filename:
            try:
                self.getParameters().save(filename)
            except Exception as e:
                self.showDetailedErrorMessage(
                        "Error saving settings to "+filename+": "+repr(e),
                        traceback.format_exc())

    def setParameters(self, p):
        """Copy from some kind of parameters object to UI."""
        self.fileLoader.loadParameters(p)
        self.imageVisualizer.loadParameters(p)

        # self.lineEditDirectory.setText(p.directory)
        # self.comboBoxDirectory.setCurrentIndex(p.dirMode)
        # self.checkBoxRdcAutosave.setChecked(p.autosave)

        self.comboBoxAlgorithm.setCurrentIndex(
            self.dcAlgorithmNames.index(p.dcAlgorithm))
        self.checkBoxImpute.setChecked(p.dcImpute)
        self.comboBoxDerivative.setCurrentIndex(p.dcDerivative)
        self.spinBoxDerivativeWindow.setValue(p.dcDerivativeWindow)
        self.spinBoxDerivativePoly.setValue(p.dcDerivativePoly)
        self.spinBoxComponents.setValue(p.dcComponents)
        self.comboBoxStartingPoint.setCurrentIndex(p.dcStartingPoint)
        self.comboBoxInitialValues.setCurrentIndex(
            self.dcInitialValuesNames.index(p.dcInitialValues))
        self.lineEditInitSkew.setValue(p.dcInitialValuesSkew)
        self.spinBoxInitPower.setValue(p.dcInitialValuesPower)
        self.lineEditInitNoise.setValue(p.dcInitialValuesNoise)
        self.comboBoxUseRoi.setCurrentIndex(self.useRoiNames.index(p.dcRoi))
        self.checkBoxContrast.setChecked(p.dcContrast)
        self.comboBoxContrast.setCurrentIndex(p.dcContrastConcentrations)
        self.lineEditContrast.setValue(p.dcContrastWeight)
        self.spinBoxIterations.setValue(p.dcIterations)
        self.lineEditTolerance.setValue(p.dcTolerance)

        self.comboBoxClusterInput.setCurrentIndex(
            self.caInputNames.index(p.caInput))
        self.comboBoxClusterNormalization.setCurrentIndex(
             self.caNormalizationNames.index(p.caNormalization))
        self.comboBoxClusterMethod.setCurrentIndex(
            self.caMethodNames.index(p.caMethod))
        self.spinBoxClusterClusters.setValue(p.caClusters)
        self.spinBoxClusterRestarts.setValue(p.caRestarts)
        annots = set()
        for i in range(self.listWidgetClusterAnnotations.count()):
            item = self.listWidgetClusterAnnotations.item(i)
            annots.add(item.text())
        for a in p.caAnnotations:
            if a not in annots:
                self.caAddAnnotation(name=a, edit=False)


    def loadParameters(self, checked=False, filename=None):
        """
        Load parameters from a file, showing a dialog if no filename is given.
        The 'checked' parameter is just a placeholder to match QPushButton.clicked().
        """
        if not filename:
            filename = self.getLoadFileName(
                    "Load decomposition settings",
                    filter="Settings files (*.djs);;All files (*)",
                    settingname='settingsDir')
        if filename:
            p = self.getParameters()
            try:
                p.load(filename)
                self.setParameters(p)
            except Exception as e:
                self.showDetailedErrorMessage(
                        "Error loading settings from "+filename+": "+repr(e),
                        traceback.format_exc())



def main():
    parser = argparse.ArgumentParser(
            description='Graphical application for decomposition of '+
            'hyperspectral data.')
    parser.add_argument('files', metavar='file', nargs='*',
                        help='initial hyperspectral images to load')
    parser.add_argument('-p', '--params', metavar='file.djs', dest='paramFile',
                        help='parameter file to load')
    MyMainWindow.run_octavvs_application(
        parser=parser, parameters=['files', 'paramFile'])





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 18 14:45:53 2021

@author: carl
"""

import numpy as np
import scipy
import sklearn
import sklearn.metrics
import time

def simplisma(d, nr, f):
    """
    The SIMPLISMA algorithm for finding a set of 'pure' spectra to serve
    as starting point for MCR-ALS etc.
    Reference Matlab Code:
        J. Jaumot, R. Gargallo, A. de Juan, R. Tauler,
        Chemometrics and Intelligent Laboratoty Systems, 76 (2005) 101-110

    Parameters
    ----------
    d : array(nspectra, nwavenums)
        input spectra.
    nr : int
        number of output components.
    f : float
        noise threshold.

    Returns
    -------
    spout: array(nr, nspectra)
        concentration profiles of 'purest' spectra.
    imp : array(nr, dtype=int)
        indexes of the 'purest' spectra.
    """

    nrow = d.shape[0]
    ncol = d.shape[1]
    s = d.std(axis=0)
    m = d.mean(axis=0)
    mf = m + m.max() * f
    p = s / mf

    # First Pure Spectral/Concentration profile
    imp = np.empty(nr, dtype=np.int)
    imp[0] = p.argmax()

    #Calculation of correlation matrix
    l2 = s**2 + mf**2
    dl = d / np.sqrt(l2)
    c = (dl.T @ dl) / nrow

    #calculation of the first weight
    w = (s**2 + m**2) / l2
    p *= w
    #calculation of following weights
    dm = np.zeros((nr+1, nr+1))
    for i in range(1, nr):
        dm[1:i+1, 1:i+1] = c[imp[:i], :][:, imp[:i]]
        for j in range(ncol):
            dm[0, 0] = c[j, j]
            dm[0, 1:i+1] = c[j, imp[:i]]
            dm[1:i+1, 0] = c[imp[:i], j]
            w[j] = np.linalg.det(dm[0:i+1, 0:i+1])
        imp[i] = (p * w).argmax()

    ss = d[:,imp]
    spout = ss / np.sqrt(np.sum(ss**2, axis=0))
    return spout.T, imp

# import pymcr


# def pymcr_als(sp, initial_components, maxiters, reltol,
#             callback_iter):

#     mcr = pymcr.mcr.McrAR(max_iter=maxiters,
#                           tol_err_change=reltol,
#                           tol_increase=1., tol_n_increase=10,
#                           tol_n_above_min=30)
#     mcr.fit(sp, ST=initial_components, post_iter_fcn=callback_iter)

#     return mcr.n_iter, mcr.C_opt_.T, mcr.ST_opt_, np.asarray(mcr.err)


def mcr_als(sp, initial_components, maxiters, nonnegative=(True, True),
            tol_abs_error=0, tol_rel_improv=None, tol_ups_after_best=None,
            maxtime=None, callback=None, acceleration=None,
            return_time=False, **kwargs):
    """
    Perform MCR-ALS nonnegative matrix decomposition on the matrix sp

    Parameters
    ----------
    sp : array(nsamples, nfeatures)
        Spectra to be decomposed.
    initial_components : array(ncomponents, nfeatures)
        Initial concentrations.
    maxiters : int
        Maximum number of iterations.
    nonnegative : pair of bool
        True if (initial, other) components must be non-negative
    tol_abs_error : float, optional
        Error target (mean square error).
    tol_rel_improv : float, optional
        Stop when relative improvement is less than this over 10 iterations.
    tol_ups_after_best : int, optional
        Stop after error going net up this many times since best error.
    maxtime : float, optional
        Stop after this many seconds of process time have elapsed
    callback : func(it : int, err : float, A : array, B : array)
        Callback for every iteration.
    acceleration : str
        None, 'Anderson', 'AdaptiveOverstep'.
        Use one of two stabilized acceleration schemes.
        Anderson acceleration operates on whole iterations (A+B updates),
        mixing earlier directions to step towards the fixed point. This
        implementation temporarily falls back to basic updates if those
        would be much better.
        The AdaptiveOverstep algorithm attempts to gradually modify the
        step length, picking the best of a shorter or longer step.
    m : int, >1
        For Anderson acceleration: the number of earlier steps to consider.
    return_time : bool
        Measure and return process_time at each iteration.

    Returns
    -------
    A : array(ncomponents, nfeatures)
        Spectra (at lowest error)
    B : array(ncomponents, nsamples)
        Concentrations at lowest error
    error : list(float)
        Mean square error at every iteration
    process_time : list(float)
        Time relative start at each iteration, only if return_time is True.
    """
    nrow, ncol = np.shape(sp)
    nr = initial_components.shape[0]
    A = initial_components.T.copy()
    B = np.empty((nr, nrow))
    errors = []
    errorbest = None # Avoid spurious warning
    prevA, prevB = (None, None)

    unknown_args = kwargs.keys() - {'m', 'aosettings'}
    if unknown_args:
        raise TypeError('Unknown arguments: {}'.format(unknown_args))

    if acceleration == 'AndersonOld':
        ason_m = kwargs.get('m', 2)
        ason_g = None
        ason_G = []
        ason_X = []
    elif acceleration == 'Anderson':
        ason_m = kwargs.get('m', 2)
        ason_g = None
        ason_G = []
        ason_X = []
    elif acceleration:
        raise ValueError("acceleration must be None or 'Anderson'")

    starttime = time.process_time()
    if return_time:
        times = []
    tol_rel_iters = 10

    aerror = berror = saveA = 0 # Avoid warnings
    for it in range(maxiters):
        # print('strides', it, A.strides, B.strides)
        # Update B
        while(True):
            prevB = B
            if nonnegative[1]:
                error = 0
                B = np.empty_like(B)
                for i in range(nrow):
                    B[:, i], res = scipy.optimize.nnls(A, sp[i, :])
                    error = error + res * res
            else:
                B, res, _, _ = np.linalg.lstsq(A, sp.T, rcond=-1)
                error = res.sum()

            if acceleration == 'Anderson' and len(ason_X) > 1 \
                and error > aerror:
                print('restart', it, error-aerror, error-berror)
                print('gamma', gamma, np.linalg.norm(ason_g))
                ason_X = []
                ason_G = []
                A = saveA
            else:
                if acceleration == 'Anderson' and len(ason_X) < 1:
                    print('Bstep', it, error-aerror, error-berror)
                break
        berror = error

        prevA = A
        if nonnegative[0]:
            error = 0
            A = np.empty_like(A)
            for i in range(ncol):
                A[i, :], res = scipy.optimize.nnls(B.T, sp[:, i])
                error = error + res * res
        else:
            A, res, _, _ = np.linalg.lstsq(B.T, sp, rcond=-1)
            A = A.T
            error = res.sum()

        if acceleration == 'Anderson':
            aerror = error
            prevg = ason_g
            ason_g = (A - prevA).flatten()
            if len(ason_X) < 1:
                ason_X.append(ason_g)
            else:
                ason_G.append(ason_g - prevg)
                while(len(ason_G) > ason_m):
                    ason_G.pop(0)
                    ason_X.pop(0)
                Garr = np.asarray(ason_G)
                try:
                    gamma = np.linalg.lstsq(Garr.T, ason_g, rcond=-1)[0]
                    # print('gamma', gamma, np.linalg.norm(ason_g))
                except np.linalg.LinAlgError:
                    print('lstsq failed to converge; restart at iter %d' % it)
                    print('nans', np.isnan(Garr).sum(), np.isnan(ason_g).sum())
                    ason_X = []
                    ason_G = []
                else:
                    dx = ason_g - gamma @ (np.asarray(ason_X) + Garr)
                    ason_X.append(dx)
                    saveA = A
                    # A = prevA + dx.reshape(A.shape)
                    if nonnegative[0]:
                        A = np.maximum(0, A)

        error = error / sp.size

        # Anderson
        # A0 -> B0   errb0
        # B0 -> A1   erra1:E  A1-A0: X0,g0
        # A1 -> B1   errb1
        # B1 -> A2   erra2:E  A2-A1: g1  g1-g0: G0   g1=v0G: v0
        #            g1-v0(X+G): X1   A1+X1: A'2
        # A'2 -> B2  errb2  if errb2>erra2:  XG=[]  A2 -> B2   (seed??)
        # B2 -> A3   erra3:E  A3-A2: g2  g2-g1: G1   g2=v1G: v1
        #            g2-v1(X+G): X2   A2+X2: A'3
        # A'3 -> B3  errb3  if errb3>erra3:  XG=[]  A3 -> B3   (seed??)

        if acceleration == 'AndersonOld' and it > 2:
            prevg = ason_g
            ason_g = np.empty(A.size + B.size)
            ason_g[:A.size] = (A - prevA).flatten()
            ason_g[A.size:] = (B - prevB).flatten()
            if len(ason_X) < 1:
                ason_X.append(ason_g)
            else:
                ason_G.append(ason_g - prevg)
                while(len(ason_G) > ason_m):
                    ason_G.pop(0)
                    ason_X.pop(0)
                Garr = np.asarray(ason_G)
                try:
                    gamma = np.linalg.lstsq(Garr.T, ason_g, rcond=-1)[0]
                except np.linalg.LinAlgError:
                    print('lstsq failed to converge; restart at iter %d' % it)
                    print('nans', np.isnan(Garr).sum(), np.isnan(ason_g).sum())
                    ason_X = []
                    ason_G = []
                else:
                    # print('gamma', gamma, np.linalg.norm(ason_g))
                    xstep = ason_g - gamma @ (np.asarray(ason_X) + Garr)
                    ason_X.append(xstep)
                    nA = prevA + xstep[:A.size].reshape(A.shape)
                    if nonnegative[0]:
                        nA = np.maximum(0, nA)
                    nB = prevB + xstep[A.size:].reshape(B.shape)
                    if nonnegative[1]:
                        nB = np.maximum(0, nB)
                    nerror = np.linalg.norm(sp - nB.T @ nA.T)**2 / sp.size
                    # Reject changes that are much worse than the basic step
                    de = error - errors[-1]
                    nde = nerror - errors[-1]
                    if nde > .5 * de or (de > 0 and nde > 2 * de):
                        ason_X = []
                        ason_G = []
                        print('aold restart',it)
                    else:
                        A, B, error = (nA, nB, nerror)


        curtime = time.process_time() - starttime
        if return_time:
            times.append(curtime)
        errors.append(error)
        if not it or error < errorbest:
            errorbest = error
            Abest = A
            Bbest = B
            netups = 0
        if it:
            if error < tol_abs_error:
                break
            if tol_rel_improv and it > tol_rel_iters:
                emax = max(errors[-tol_rel_iters-1:-2])
                if (emax - errors[-1]) * tol_rel_iters <= \
                    tol_rel_improv * emax:
                    break
            if tol_ups_after_best is not None:
                if error < errors[-2]:
                    netups = max(0, netups - 1)
                else:
                    netups = netups + 1
                    if netups > tol_ups_after_best:
                        break
        if it and maxtime and curtime >= maxtime:
            break
        if callback is not None:
            callback(it, errors, A.T, B)

    if return_time:
        return Abest.T, Bbest, errors, times
    return Abest.T, Bbest, errors




#from PyQt5 import QtWidgets
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as Canvas
from PyQt5.QtCore import pyqtSignal, pyqtSlot
import numpy as np



class MplWidget(Canvas):
    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        self.canvas = self
        Canvas.__init__(self, self.fig)

    def Invert(self):
        # print('called')
        self.ax.invert_xaxis()


class MplWidgetC(Canvas):
    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off() 
        self.ax.set_position([0,0,1,1])
        self.canvas = self
        Canvas.__init__(self, self.fig)
 
   


class Mpl_Proj(MplWidget):
    def __init__(self, parent=None):
        MplWidget.__init__(self, parent)
        self.image = None
        self.artists = []
        self.points = None
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        self.image = None

    @pyqtSlot(np.ndarray, str)
    def setImage(self, img, cmap):
#        self.ax.clear()
#        self.artists = []
#        self.fig.tight_layout()
#        self.draw_idle()
        if self.image is None:
            self.image = self.ax.imshow(img, cmap)
        else:
            self.image.set_data(img)
            self.image.set_extent((0, img.shape[1], img.shape[0], 0))
            self.ax.set_xlim(0, img.shape[1])
            self.ax.set_ylim(img.shape[0], 0)
            self.setCmap(cmap)
        self.image.autoscale()

    @pyqtSlot(str)
    def setCmap(self, cmap):
        self.image.set_cmap(cmap)
        self.draw_idle()

    @pyqtSlot()
    def clearMarkings(self):
        for a in self.artists:
            a.remove()
        self.artists = []
        if self.points is not None:
            self.points.remove()
            self.points = None

    @pyqtSlot(list)
    def setRoi(self, pts):
        self.clearMarkings()
        if len(pts) > 1:
            npts = np.array(pts + [pts[0]]).T
            self.artists = self.ax.plot(npts[0]+.5, npts[1]+.5, 'r')
        self.draw_idle()

    def addPoints(self, pts):
        if self.points is not None:
            self.points.remove()
        self.points = self.ax.scatter(pts.T[0,:], pts.T[1,:], marker='p', color='black')
        self.draw_idle()


class MPL_ROI(Canvas):

    captured = pyqtSignal(list)

    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=False)
        self.ax = self.fig.add_subplot(111)
        self.ax.set_axis_off()
        self.ax.set_position([0,0,1,1])
        Canvas.__init__(self, self.fig)
        self.mpl_connect('button_press_event', self.onclick)

        self.points = []
        self.artists = []
        self.wh = (1,1)
        self.mainimg = self.ax.imshow([[0]], cmap='gray', zorder=0, aspect='equal')

    def setImage(self, img, cmap=None):
        self.mainimg.set_data(img)
        self.mainimg.set_cmap(cmap)
        self.mainimg.autoscale()
        self.mainimg.set_extent((0, self.wh[1], self.wh[0], 0))
        self.ax.autoscale(False)
        self.ax.set_position([0,0,1,1])
        self.ax.set_xlim(0, self.wh[1])
        self.ax.set_ylim(self.wh[0], 0)
        self.draw_idle()

    def setCmap(self, cmap):
        self.mainimg.set_cmap(cmap)
        self.draw_idle()

    def setSize(self, wh):
        if wh != self.wh:
            self.resetAll()
        self.wh = wh


    def onclick(self, event):
        x, y = event.xdata, event.ydata
        if (x is None) or (y is None):
            return

        x = int(x)
        y = int(y)
        if not (0 <= x < self.wh[1] and 0 <= y < self.wh[0]):
            print('click outside',x,y,self.wh)
            return
        if not len(self.points):
            self.artists.append(self.ax.plot(x+.5, y+.5, 'y*'))
        else:
            self.artists.append(self.ax.plot((self.points[-1][0]+.5, x+.5), (self.points[-1][1]+.5, y+.5), 'r-'))

        self.points.append((x, y))
        self.draw_idle()
        self.captured.emit(self.points)

    @pyqtSlot()
    def resetAll(self):
        if len(self.points):
            self.points = []
            for la in self.artists:
                for a in la:
                    a.remove()
            self.artists = []
            self.draw_idle()
        self.captured.emit(self.points)

    @pyqtSlot()
    def removeOne(self):
        if len(self.points):
            self.points.pop()
            for a in self.artists.pop():
                a.remove()
            self.draw_idle()

        self.captured.emit(self.points)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 18 14:45:53 2021

@author: carl
"""

import numpy as np
import scipy
import math
import time
from threadpoolctl import threadpool_limits

def simplisma(d, nr, f):
    """
    The SIMPLISMA algorithm for finding a set of 'pure' spectra to serve
    as starting point for MCR-ALS etc.
    Reference Matlab Code:
        J. Jaumot, R. Gargallo, A. de Juan, R. Tauler,
        Chemometrics and Intelligent Laboratoty Systems, 76 (2005) 101-110

    Parameters
    ----------
    d : array(nspectra, nwavenums)
        input spectra.
    nr : int
        number of output components.
    f : float
        noise threshold.

    Returns
    -------
    spout: array(nr, nspectra)
        concentration profiles of 'purest' spectra.
    imp : array(nr, dtype=int)
        indexes of the 'purest' spectra.
    """

    nrow = d.shape[0]
    ncol = d.shape[1]
    s = d.std(axis=0)
    m = d.mean(axis=0)
    mf = m + m.max() * f
    p = s / mf

    # First Pure Spectral/Concentration profile
    imp = np.empty(nr, dtype=np.int)
    imp[0] = p.argmax()

    #Calculation of correlation matrix
    l2 = s**2 + mf**2
    dl = d / np.sqrt(l2)
    c = (dl.T @ dl) / nrow

    #calculation of the first weight
    w = (s**2 + m**2) / l2
    p *= w
    #calculation of following weights
    dm = np.zeros((nr+1, nr+1))
    for i in range(1, nr):
        dm[1:i+1, 1:i+1] = c[imp[:i], :][:, imp[:i]]
        for j in range(ncol):
            dm[0, 0] = c[j, j]
            dm[0, 1:i+1] = c[j, imp[:i]]
            dm[1:i+1, 0] = c[imp[:i], j]
            w[j] = np.linalg.det(dm[0:i+1, 0:i+1])
        imp[i] = (p * w).argmax()

    ss = d[:,imp]
    spout = ss / np.sqrt(np.sum(ss**2, axis=0))
    return spout.T, imp

def clustersubtract(data, components, skewness=100, power=2):
    """
    Create initial spectra for MCR-ALS based on successively removing
    what appears to be the strongest remaining component.

    Parameters
    ----------
    data : array (nspectra, nfeatures)
        Spectral data.
    components : int
        Number of components to return.
    skewness : float, optional
        Asymmetry between positive and negative residuals when computing
        how much of the previous component to remove from the data.
        The default is 100.
    power : float, optional
        The sum of residuals is raised to this power before summation to
        determine the leading remaining component.

    Returns
    -------
    initial_spectra : array (components, nfeatures)
    """
    def typical_cluster(data, first):
        # draw sqrt(n) random data points, r
        # find closest in r for each s in data
        # for r with most s, return mean of those s
        r = np.random.choice(len(data), math.floor(math.sqrt(len(data))))
        rd = data[r]
        nearest = scipy.spatial.distance.cdist(
            rd, data, 'cosine').argmin(axis=0)
        # Mean of those who are nearest the biggest cluster
        if first:
            selected = np.bincount(nearest).argmax()
        else:
            sums = data.sum(1) ** power
            selected = np.bincount(nearest, weights=sums).argmax()
        return data[nearest == selected].mean(0)

    comps = []
    for c in range(components):
        tc = typical_cluster(data, c == 0)
        tc = np.maximum(tc, 0)
        tc = tc / (tc * tc).sum() ** .5

        comps.append(tc)
        sgn = np.ones_like(data, dtype=bool)
        for i in range(10):
            ww = 1 * sgn + skewness * ~sgn
            a = (data * ww * tc).sum(1) / (ww * tc * tc).sum(1)
            oldsgn = sgn
            sgn = data > a[:, None] @ tc[None, :]
            if np.array_equal(sgn, oldsgn):
                break
        data = data - a[:, None] @ tc[None, :]
    return np.array(comps)

def numpy_scipy_threading_fix_(func):
    """
    This decorator for mcr_als prevents threading in BLAS if scipy's NNLS
    is used, because for some reason NNLS won't be parallelized if called
    shortly after lstsq or @. This makes a *massive* difference to the
    time needed for Anderson acceleration, where the BLAS calls themselves
    take negligible time. For mixed NNLS/lstsq solving (of MCR-ALS on
    derivatives) it's less obvious whether NNSL or lstsq should be allowed
    to be parallelized.
    Note: This issue is seen on my Linux setup and might be highly version-
    dependent. More investigation is needed.
    """
    def check(*args, **kwargs):
        if 'nonnegative' not in kwargs or np.any(kwargs['nonnegative']):
            with threadpool_limits(1, 'blas'):
                return func(*args, **kwargs)
        else:
            return func(*args, **kwargs)
    return check

@numpy_scipy_threading_fix_
def mcr_als(sp, initial_A, *, maxiters, nonnegative=(True, True),
            tol_abs_error=0, tol_rel_improv=None, tol_iters_after_best=None,
            maxtime=None, callback=None, acceleration=None, normalize=None,
            contrast_weight=None, return_time=False, **kwargs):
    """
    Perform MCR-ALS nonnegative matrix decomposition on the matrix sp

    Parameters
    ----------
    sp : array(nsamples, nfeatures)
        Spectra to be decomposed.
    initial_A : array(ncomponents, nfeatures)
        Initial spectra or concentrations.
    maxiters : int
        Maximum number of iterations.
    nonnegative : pair of bool, default (True, True)
        True if (initial, other) components must be non-negative
    tol_abs_error : float, optional
        Error target (mean square error).
    tol_rel_improv : float, optional
        Stop when relative improvement is less than this over 10 iterations.
    tol_iters_after_best : int, optional
        Stop after this many iterations since last best error.
    maxtime : float, optional
        Stop after this many seconds of process time have elapsed
    callback : func(it : int, err : float, A : array, B : array)
        Callback for every iteration.
    acceleration : str, optional
        None or 'Anderson'.
        Anderson acceleration operates on whole iterations (A or B updates),
        mixing earlier directions to step towards the fixed point. This
        implementation restarts from basic updates when those would be
        better.
    normalize : str, optional
        Which matrix to l2 normalize: None, 'A' or 'B'
    contrast_weight : (str, float), optional
        Increase contrast in one matrix by mixing the other, named matrix
        ('A' or 'B') with the mean of its vectors. If A is spectra,
        try contrast_weight=('B', 0.05) to increase spectral contrast.
        See Windig and Keenan, Applied Spectroscopy 65: 349 (2011).
    return_time : bool, default False
        Measure and return process_time at each iteration.

    Anderson acceleration parameters in kwargs
    -------
    m : int, >1, default 2
        The maximum number of earlier steps to consider.
    alternate : bool, default True
        Alternate between accelerating A and B, switching when restarting.
    beta : float, default 1.
        Scaling factor for accelerated step length.
    betascale : float, default 1.
        Reduction factor for beta after each restart.
    bmode : bool, default False
        Start with accelerating B instead of A.

    Returns
    -------
    A : array(ncomponents, nfeatures)
        Spectra (at lowest error)
    B : array(ncomponents, nsamples)
        Concentrations at lowest error
    error : list(float)
        Mean square error at every iteration
    process_time : list(float)
        Time relative start at each iteration, only if return_time is True.
    """
    if normalize not in [None, 'A', 'B']:
        raise ValueError('Normalization must be None, A or B')
    unknown_args = kwargs.keys() - {
        'm', 'alternate', 'beta', 'betascale', 'bmode'}
    if unknown_args:
        raise TypeError('Unknown arguments: {}'.format(unknown_args))

    nrow, ncol = sp.shape
    nr = initial_A.shape[0]
    if normalize == 'A':
        norm = np.linalg.norm(initial_A, axis=1)
        A = np.divide(initial_A.T, norm, where=norm!=0,
                      out=np.zeros(initial_A.shape[::-1]))
    else:
        A = initial_A.T.copy()
    B = np.empty((nr, nrow))
    errors = []
    errorbest = None # Avoid spurious warning
    # prevA, prevB = (None, None)
    newA = newB = None
    error = preverror = None

    cw = 0
    if contrast_weight is not None:
        if contrast_weight[0] == 'A':
            cw = contrast_weight[1]
        elif contrast_weight[0] == 'B':
            cw = -contrast_weight[1]
        else:
            raise ValueError("contrast_weight must be ('A'|'B', [0-1])")


    if acceleration == 'Anderson':
        ason_Bmode = kwargs.get('bmode', False)
        ason_alternate = kwargs.get('alternate', True)
        ason_m = kwargs.get('m', 2)
        ason_beta = kwargs.get('beta', 1.)
        ason_betascale = kwargs.get('betascale', 1.)
        ason_g = None
        ason_G = []
        ason_X = []
    elif acceleration:
        raise ValueError("acceleration must be None or 'Anderson'")

    starttime = time.process_time()
    if return_time:
        times = []
    tol_rel_iters = 10

    for it in range(maxiters):
        ba = 0
        retry = False
        while ba < 2:
            if not retry:
                preverror = error
            if ba == 0:
                if newA is None:
                    newA = A
                prevA = newA
                if cw > 0:
                    newA = (1 - cw) * newA + cw * newA.mean(1)[:,None]
                if nonnegative[1]:
                    error = 0
                    if not retry:
                        B = np.empty_like(B)
                    for i in range(nrow):
                        B[:, i], res = scipy.optimize.nnls(newA, sp[i, :])
                        error = error + res * res
                else:
                    B, res, _, _ = np.linalg.lstsq(newA, sp.T, rcond=-1)
                    error = res.sum()
                if normalize == 'B':
                    norm = np.linalg.norm(B, axis=1)
                    B = np.divide(B.T, norm, where=norm!=0, out=B.T).T
                newA = None
            else:
                if newB is None:
                    newB = B
                prevB = newB
                if cw < 0:
                    newB = (1 + cw) * newB - cw * newB.mean(1)[:,None]
                if nonnegative[0]:
                    error = 0
                    if not retry:
                        A = np.empty_like(A)
                    for i in range(ncol):
                        A[i, :], res = scipy.optimize.nnls(newB.T, sp[:, i])
                        error = error + res * res
                else:
                    A, res, _, _ = np.linalg.lstsq(newB.T, sp, rcond=-1)
                    A = A.T
                    error = res.sum()
                if normalize == 'A':
                    norm = np.linalg.norm(A, axis=0)
                    np.divide(A, norm, where=norm!=0, out=A)
                newB = None

            if acceleration is None:
                pass
            elif ba == ason_Bmode:
                if retry:
                    retry = False
                    if ason_alternate:
                        ason_Bmode = not ason_Bmode
                    ason_beta = ason_beta * ason_betascale
                elif len(ason_X) > 1 and error > preverror:
                    ason_X = []
                    ason_G = []
                    retry = True
                    ba = ba - 1
                else:
                    pass
            elif ason_Bmode == 1 and it == 0:
                pass
            else:
                prevg = ason_g
                ason_g = ((A - prevA) if ba else (B - prevB)).flatten()
                if len(ason_X) < 1:
                    ason_X.append(ason_g)
                else:
                    ason_G.append(ason_g - prevg)
                    while(len(ason_G) > ason_m):
                        ason_G.pop(0)
                        ason_X.pop(0)
                    Garr = np.asarray(ason_G)
                    try:
                        gamma = scipy.linalg.lstsq(Garr.T, ason_g)[0]
                    except scipy.linalg.LinAlgError:
                        print('lstsq failed to converge; '
                              'restart at iter %d' % it)
                        # print('nans', np.isnan(Garr).sum(),
                        #       np.isnan(ason_g).sum())
                        ason_X = []
                        ason_G = []
                    else:
                        gamma = ason_beta * gamma
                        dx = ason_g - gamma @ (np.asarray(ason_X) + Garr)
                        ason_X.append(dx)
                        if ba:
                            newA = prevA + dx.reshape(A.shape)
                            if nonnegative[0]:
                                np.maximum(0, newA, out=newA)
                        else:
                            newB = prevB + dx.reshape(B.shape)
                            if nonnegative[1]:
                                np.maximum(0, newB, out=newB)
            ba = ba + 1
        # error = error / sp.size

        curtime = time.process_time() - starttime
        if return_time:
            times.append(curtime)
        errors.append(error)
        if not it or error < errorbest:
            errorbest = error
            Abest = A
            Bbest = B
            iterbest = it
        if it:
            if error < tol_abs_error:
                break
            if tol_rel_improv and it > tol_rel_iters:
                emax = max(errors[-tol_rel_iters-1:-2])
                if (emax - errors[-1]) <= \
                    tol_rel_improv * tol_rel_iters * emax:
                        break
            if tol_iters_after_best is not None:
                if iterbest + tol_iters_after_best < it:
                    break
        if it and maxtime and curtime >= maxtime:
            break
        if callback is not None:
            callback(it, errors, A.T, B)

    if return_time:
        return Abest.T, Bbest, errors, times
    return Abest.T, Bbest, errors




"""
Compilation of function for the ftir main
"""

import numpy as np
import scipy.io as sio
#from scipy.signal import savgol_filter, wiener
#import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
#from skimage import filters
#from sklearn.cluster import SpectralClustering, KMeans
#from scipy.sparse import spdiags, csc_matrix
#from scipy.sparse.linalg import spsolve
#from scipy import sparse
#from scipy.spatial import ConvexHull
#from scipy.interpolate import interp1d
#import scipy as sc




"""
#######################read data from matlab input file#######################
"""
def readmat(filename):
     s = sio.loadmat(filename)
     info = sio.whosmat(filename)
     ss = s[(str(info[0][0]))]
     wavenumber=ss[:,0]
     sizex = len(wavenumber)
     sizemx, sizemy = np.shape(ss)
     sp = ss[:,1:]
     if (len(info)) > 1:
         (l,*_) = s['wh']
         w , h = l
     else:      
         w = int(np.sqrt(sp.shape[1]))
         h = sp.shape[1] // w
         if w * h != sp.shape[1]:
             w = sp.shape[1]
             h = 1
     p = sp.reshape(sizex,h,w,order='C')
     return  w, h, p, wavenumber, sp


#(1) Based on area under spectrum
def proarea(d3image,x):
    i,j,k = np.shape(d3image)
    x = np.sort(x) #because data are scanned from high to low
    cc=np.zeros((j,k))
    for ii in range(0,j):
         for jj in range(0,k):
              cc[ii,jj]= np.trapz(d3image[:,ii,jj],x)
    return cc

#area of single spectra
def area(y):
    area = np.trapz(y)
    return area

#(2) Based on the wavenumber (cm-1)
def prowavenum(d3image,wavenums,wavenumv):
    i,j,k = np.shape(d3image)
    cc=np.zeros((j,k))
    idx = (np.abs(wavenums-wavenumv)).argmin()
    cc = d3image[idx,:,:]
    return cc

#(3) Based on Maximum Intensity Projection
def promip(d3image):
    i,j,k = np.shape(d3image)
    brp = d3image[:,:,:]
    cmode1 = brp.max(0) #max
    return cmode1

#(4) Based on local area under curve
#prerequisite Gaussian Function
def proloa(d3image,wavenum,wavenumv):
    i,j,k = np.shape(d3image)
    cc=np.zeros((j,k))
#    uu=np.zeros((j,k))
    x = wavenum[::-1]
    i1 = (np.abs(x-(wavenumv-100))).argmin()
    i2 = (np.abs(x-(wavenumv+100))).argmin()
    init_vals = [1, 2000, 500]  # for [amp, cen, wid]
    for ii in range(0,j):
        for jj in range(0,k):
#            uu= np.trapz(d3image[:,ii,jj],x)
            spectra = d3image[:,ii,jj]
            spectra = spectra[::-1]
            yplot = spectra[i1:i2+1]
            xplot = x[i1:i2+1]
            best_vals, covar = curve_fit(gaussian, xplot, yplot, p0=init_vals)
            cc[ii,jj]= np.trapz(gaussian(xplot, *best_vals),xplot)#/uu
 #           cc[ii,jj]= np.trapz(yplot,xplot)
    return cc


"""
Function to evaluat the initial value using SIMPLISMA
"""
def initi_simplisma(d,nr,f):
    """
    Function to calculte the pure profiles
    Reference Matlab Code:
         J. Jaumot, R. Gargallo, A. de Juan, R. Tauler,
         Chemometrics and Intelligent Laboratoty Systems, 76 (2005) 101-110


    ---:
        input : float d(nrow,ncol) = original spectrum
                integer nr = the number of pure components
                float f = number of noise allowed eg. 0.1 (10%)

        output: float spout(nr,nrow) = purest number component profile
                integer imp(nr) = indexes of purest spectra
    """
    nrow = d.shape[0]
    ncol = d.shape[1]

    s = d.std(axis=0)
    m = d.mean(axis=0)
    mf = m + m.max() * f
    p = s / mf

    # First Pure Spectral/Concentration profile
    imp = np.empty(nr, dtype=np.int)
    imp[0] = p.argmax()

    #Calculation of correlation matrix
    l2 = s**2 + mf**2
    dl = d / np.sqrt(l2)
    c = (dl.T @ dl) / nrow

    #calculation of the first weight
    w = (s**2 + m**2) / l2
    p *= w
    #calculation of following weights
    dm = np.zeros((nr+1, nr+1))
    for i in range(1, nr):
        dm[1:i+1,1:i+1] = c[imp[:i],:][:,imp[:i]]
        for j in range(ncol):
            dm[0,0] = c[j,j]
            dm[0,1:i+1]=c[j,imp[:i]]
            dm[1:i+1,0]=c[imp[:i],j]
            w[j] = np.linalg.det(dm[0:i+1, 0:i+1])

        imp[i] = (p * w).argmax()

    ss = d[:,imp]
    spout = ss / np.sqrt(np.sum(ss**2, axis=0))
    return spout.T, imp



def listindexes(l):
    r = {}
    for i in range(len(l)):
        v = l[i]
        if v in r:
            r[v].append(i)
        else:
            r[v] = [i]
    return r



from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
#import matplotlib.patches as patches
import numpy as np
#import random



class DataPlotWidget(FigureCanvas):
    updated = pyqtSignal()
    clicked = pyqtSignal()

    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
#        self.fig.canvas.mpl_connect('button_press_event', self.onclick)
        FigureCanvas.setSizePolicy(self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)
        self.wavenumber = None
        self.inspectra = None
        self.spectra = None
        self.popax = None
        self.window_title = self.__class__.__name__

    def mousePressEvent(self, event):
        self.clicked.emit()

    def setData(self, wavenumber, inspectra, spectra):
        if (self.wavenumber is wavenumber and
            self.inspectra is inspectra and
            self.spectra is spectra):
            return
        # print('update',self.__class__.__name__,
        #       self.inspectra.shape if self.inspectra is not None else None,
        #       self.spectra.shape if self.spectra is not None else None)
        # Skip redraw if there is and was no outdata
        if self.spectra is not None or spectra is not None:
            self.draw_idle()
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.updated.emit()

    def getSpectra(self):
        return self.spectra if self.spectra is not None else self.inspectra

    def getWavenumbers(self):
        return self.wavenumber

    def draw_ax(self, ax):
        pass

    def draw(self):
        self.ax.clear()
        if self.spectra is not None and len(self.spectra) > 0:
            self.draw_ax(self.ax)
        FigureCanvas.draw(self)
        if plt.fignum_exists(self.objectName()):
            self.popOut()

    def popOut(self):
        fig = plt.figure(self.objectName(), tight_layout=dict(pad=.6))
        fig.canvas.set_window_title(self.window_title)
        ax = fig.gca()
        ax.clear()
        if self.spectra is not None and len(self.spectra) > 0:
            self.draw_ax(ax)
        fig.canvas.draw_idle()
        if ax != self.popax:
            fig.show()
            self.popax = ax


class SpectraPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Raw spectra'

    def draw_ax(self, ax):
        for i in self.spectra:
            ax.plot(self.wavenumber, i, linewidth=1)

class NormPlotWidget(SpectraPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Normalized spectra'

class ACPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Atmospheric correction'

    def draw_ax(self, ax):
        for s in range(len(self.spectra)):
            ax.plot(self.wavenumber, self.inspectra[s], linewidth=1,
                    label='Raw' if not s else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    color='r', linewidth=1, linestyle='--',
                    label='Atm corr.' if not s else None)
        ax.legend()


class SCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.window_title = 'Scattering correction'
        self.progressMode = False
        self.runSelected = None
        self.showSelected = None

#    def getInputSpectra(self):
#        return self.inspectra

    def setData(self, wavenumber, inspectra, spectra, runSelected, showSelected):
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.runSelected = runSelected
        self.setSelected(showSelected)

    def setSelected(self, sel):
        if self.runSelected is not None:
            if sorted(sel) == sorted(self.runSelected):
                self.showSelected = range(len(self.spectra))
            else:
                self.showSelected = []
        else:
            self.showSelected = sel
        self.draw_idle()
        self.updated.emit()

    def getSpectra(self):
        if self.spectra is None or self.runSelected is not None:
            return self.spectra
        return self.spectra[self.showSelected]

    # Rather ugly hack to show progress in the main window only: the draw method is
    # made to bypass the DataPlotWidget draw.
    def prepareProgress(self, wn):
        self.wavenumber = wn
        self.ax.clear()
        self.progressMode = True
        self.draw_idle()

    def endProgress(self):
        if self.progressMode:
            self.progressMode = False
            self.draw_idle()

    @pyqtSlot(np.ndarray, tuple)
    def progressPlot(self, curves, progress):
        if self.progressMode:
            self.ax.plot(self.wavenumber, curves.T,
                         color=plt.cm.jet(progress[0] / progress[1]), linewidth=1)
            self.draw_idle()

    def draw(self):
        if self.progressMode:
            return FigureCanvas.draw(self)
        return DataPlotWidget.draw(self)

    def draw_ax(self, ax):
        i = 0
        for s in self.showSelected:
            ax.plot(self.wavenumber, self.inspectra[s],
                    color='black', linewidth=1, label='Input' if not i else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    linewidth=1, label='Corrected' if not i else None)
            i += 1
        if i:
            ax.legend()


class SGFPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Savitzky-Golay filter'

    def draw_ax(self, ax):
        for s in range(len(self.spectra)):
            ax.plot(self.wavenumber, self.inspectra[s],
                    color='black', linewidth=1, label='Input' if not s else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    linewidth=1, label='Smoothed' if not s else None)
        if len(self.spectra):
            ax.legend()


class SRPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Spectral region selection'
        self.mincut=0
        self.maxcut=10000
        self.cutwn=np.zeros(0)

    def setData(self, wavenumber, inspectra, spectra, cutwn, mincut, maxcut):
        self.wavenumber = wavenumber
        self.inspectra = inspectra
        self.spectra = spectra
        self.mincut=mincut
        self.maxcut=maxcut
        self.cutwn=cutwn
        self.draw_idle()
        self.updated.emit()

    def getWavenumbers(self):
        return self.cutwn if self.cutwn is not None else self.wavenumber

    def draw_ax(self, ax):
        if self.spectra is None:
            print('SR draw_ax called without spectra!')
        for s in self.inspectra:
            ax.plot(self.wavenumber, s, linewidth=1)
        if self.wavenumber[-1] < self.mincut:
            ax.axvspan(self.wavenumber[-1], self.mincut, color='gray')
        if self.maxcut < self.wavenumber[0]:
            ax.axvspan(self.maxcut, self.wavenumber[0], color='gray')


class BCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'Baseline correction'

    def getSpectra(self):
        if self.spectra is not None:
            return self.inspectra - self.spectra
        return self.inspectra

    def draw_ax(self, ax):
        for s in range(len(self.spectra)):
            ax.plot(self.wavenumber, self.inspectra[s],
                    linewidth=1, label='Input' if not s else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    color='k', linewidth=1, label='Baseline' if not s else None)
#            ax.plot(self.wavenumber, self.inspectra[s] - self.spectra[s],
#                    color='gray', linewidth=1, label='Corrected' if not s else None)
            ax.legend()

class MCPlotWidget(DataPlotWidget):
    def __init__(self, parent=None):
        DataPlotWidget.__init__(self, parent=None)
        self.window_title = 'mIRage correction'

    def draw_ax(self, ax):
        for s in range(len(self.spectra)):
            ax.plot(self.wavenumber, self.inspectra[s],
                    color='black', linewidth=1, label='Input' if not s else None)
            ax.plot(self.wavenumber, self.spectra[s],
                    linewidth=1, label='Corrected' if not s else None)
        if len(self.spectra):
            ax.legend()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 14:01:08 2020

@author: carl
"""

import os.path
import numpy as np
import scipy.signal, scipy.io
from pymatreader import read_mat

from .opusreader import OpusReader
from .ptirreader import PtirReader

class SpectralData:
    """
    The list of current files and the raw data loaded from a hyperspectral
    image file in the GUIs.
    """
    def __init__(self):
        self.foldername = '' # Root dir
        self.filenames = []  # With full paths
        self.curFile = ''    # The currently loaded file
        self.wavenumber = None # array in order high->low
        self.wmin = 800
        self.wmax = 4000
        self.raw = np.empty((0,0)) # data in order (pixel, wavenumber)
        self.wh = (0, 0)  # Width and height in pixels
        self.pixelxy = None # Positions of pixels in image
        self.images = None # list of Image loaded from raw data file
        self.filetype = None # str from [txt, mat, opus, ptir]

    def set_width(self, w):
        try:
            w = int(w)
        except ValueError:
            w = 0
        if w <= 0:
            return False
        h = int(self.raw.shape[0] / w)
        if w * h != self.raw.shape[0]:
            return False
        self.wh = (w, h)
        return True

    def set_height(self, h):
        if self.set_width(h):
            self.wh = (self.wh[1], self.wh[0])
            return True
        return False

    def read_matrix(self, filename):
        """
        Read data from a file, with some error checking. The object is modified
        only if the file is successfully loaded.
        """
        wh = None
        xy = None
        images = None
        fext = os.path.splitext(filename)[1].lower()
        if fext in ['.txt', '.csv', '.mat']:
            if fext == '.mat':
                filetype = 'mat'
                try:
                    s = read_mat(filename)
                except TypeError:
                    # Workaround for uint16_codec bug (pymatreader
                    # assumes mio5, not mio4)
                    s = scipy.io.loadmat(filename)
                # Assume data are in the biggest matrix in the file
                ss = max(s.items(), key=lambda k: np.size(k[1]) )[1]
            else:
                filetype = 'txt'
                s = {}
                ss = np.loadtxt(filename)
            if 'ImR' in s and 'lambda' in s:
                ss = s['ImR']
                if ss.ndim == 2:
                    raw = ss
                else:
                    assert ss.ndim == 3
                    wh = [ss.shape[1], ss.shape[0]]
                    raw = ss.reshape((-1, ss.shape[2]))
                wn = s['lambda'].flatten()
                if wn[0] < wn[-1]:
                    wn = wn[::-1]
                    raw = raw[:, ::-1]
                assert len(wn) == raw.shape[1]
            elif ss.ndim == 2 and ss.shape[0] >= 10 and ss.shape[1] > 1:
                if 'wh' in s:
                    wh = s['wh'].flatten()
                d = -1 if ss[0,0] < ss[-1,0] else 1
                raw = ss[::d,1:].T
                wn = ss[::d,0]
            else:
                raise RuntimeError(
                    'file does not appear to describe an FTIR image matrix')
        else:
            if fext == '.ptir' or fext == '.hdf':
                filetype = 'ptir'
                reader = PtirReader(filename)
                xy = reader.xy
            else:
                filetype = 'opus'
                reader = OpusReader(filename)
            raw = reader.AB
            wn = reader.wavenum
            wh = reader.wh
            images = reader.images

        if (np.diff(wn) >= 0).any():
            raise RuntimeError('wavenumbers must be sorted')
        npix = raw.shape[0]
        if wh is not None:
            if len(wh) != 2:
                raise RuntimeError('Image size in "wh" must have length 2')
            wh = (int(wh[0]), int(wh[1]))
            if wh[0] * wh[1] != npix:
                raise RuntimeError('Image size in "wh" does not match data size')
            self.wh = wh
        elif npix != self.wh[0] * self.wh[1]:
            res = int(np.sqrt(npix))
            if npix == res * res:
                self.wh = (res, res)
            else:
                self.wh = (npix, 1)
        self.raw = raw
        self.wavenumber = wn
        self.wmin = self.wavenumber.min()
        self.wmax = self.wavenumber.max()
        self.images = images
        self.pixelxy = xy
        self.curFile = filename
        self.filetype = filetype



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jan 29 21:57:05 2020

@author: carl
"""

from PyQt5.QtWidgets import QStyle, QProxyStyle

class NoRepeatStyle(QProxyStyle):
    def styleHint(self, hint, option=None, widget=None, returnData=None):
        if hint == QStyle.SH_SpinBox_ClickAutoRepeatThreshold:
            return 10000
        else:
            return super().styleHint(hint, option, widget, returnData)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Nov 15 17:14:10 2020

@author: carl
"""

class Image:
    """
    An image, possibly white light or from a particular wavenumber
    """
    def __init__(self, data=None, name=None, fmt=None):
        self.data = data
        self.fmt = fmt
        # filename=None,
        # self.filename = filename
        self.name = name  #filename if name is None else name
        self.xy = None
        self.wh = None



from .octavvsapplication import OctavvsMainWindow
from .norepeatstyle import NoRepeatStyle

from .fileloader import FileLoader
from .imagevisualizer import ImageVisualizer

from . import constants, uitools



from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtCore import pyqtSignal, pyqtSlot
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
import matplotlib.pyplot as plt
import matplotlib
import numpy as np
import math
import skimage.draw


class BasePlotWidget(FigureCanvas):
    def __init__(self, parent=None):
        self.fig = Figure(tight_layout=dict(pad=0.4))
        self.ax = self.fig.add_subplot(111)
        FigureCanvas.__init__(self, self.fig)
        self.setParent(parent)
        FigureCanvas.setSizePolicy(
            self, QSizePolicy.Expanding, QSizePolicy.Expanding)
        FigureCanvas.updateGeometry(self)

        # 2D layout
        self.wh = None
        self.pixelxy = None
        self.pixels = 0
        # What/how to plot
        self.cmap = plt.get_cmap().name
        self.pixel_radius = 3


class RoiPlotWidget(BasePlotWidget):
    updated = pyqtSignal(int, int)  # N out of M selected

    def __init__(self, parent=None):
        super().__init__(parent)

        self.window_title = 'Region of interest'
        # Data to plot
        self.data = None
        self.roi = None
        self.roi_alpha = np.array([200, 200, 200, 128], np.uint8) # RGBA
        self.polystart = None   # Patch created when starting polygon
        self.polyline = None    # Line2D for lines in polygon
        self.drawmode = 'click' # click, add, remove
        self.drawing = False    # Drawing with the mouse?
        self.patches = None     # list of Patch for xy-pixels
        self.mpl_connect('button_press_event', self.on_click)
        self.mpl_connect('button_release_event', self.on_release)
        self.mpl_connect('motion_notify_event', self.on_move)

    def emit_updated(self):
        n = 0 if self.roi is None else self.roi.sum()
        self.updated.emit(n, self.pixels)

    def set_basic_data(self, wh=None, pixelxy=None):
        "Set shape/points and clears ROI etc"
        if wh is None and pixelxy is None:
            raise ValueError('Either wh or pixelxy must be specified')
        if pixelxy is not None:
            pixels = len(pixelxy)
        else:
            pixels = wh[0] * wh[1]
        self.wh = wh
        self.pixelxy = pixelxy
        self.pixels = pixels
        self.data = None
        self.roi = np.zeros((self.pixels), dtype=bool)
        self.polyline = None
        self.polystart = None
        self.patches = []
        self.ax.clear()
        self.emit_updated()
        self.draw_idle()

    def adjust_geometry(self, wh):
        "Update the plot geometry, keeping the pixel count"
        pixels = wh[0] * wh[1]
        assert pixels == self.pixels
        self.wh = wh
        self.set_data(self.data)

    def get_roi(self):
        "Return the ROI as a bool array (not a copy; beware)"
        return self.roi

    def set_roi(self, roi):
        "Set the ROI from an array or None"
        if roi is None:
            self.roi = np.zeros(self.pixels, dtype=bool)
        else:
            roi = np.array(roi, dtype=bool)
            if roi.shape != self.roi.shape:
                raise ValueError('ROI size must match the current geometry')
            self.roi = roi
        self.update_roi_image()
        self.emit_updated()
        self.draw_idle()

    @pyqtSlot(np.ndarray)
    def set_data(self, data):
        "Set intensities for all pixels"
        assert len(data) == self.pixels
        # (Re)create basic stuff in plot only when we get data
        if self.data is None:
            # Things common to both display modes
            self.ax.set_aspect('equal')
            self.ax.tick_params(bottom=False, labelbottom=False,
                                left=False, labelleft=False)
            self.ax.set_facecolor('#eee')

        if self.pixelxy is not None:
            cmap = plt.get_cmap(self.cmap)
            norm = matplotlib.colors.Normalize(
                vmin=data.min(), vmax=data.max())
            if self.data is None:
                minxy = np.min(self.pixelxy, axis=0)
                maxxy = np.max(self.pixelxy, axis=0)
                self.pixel_radius = (maxxy - minxy).max() * .02

                for i in range(self.pixels):
                    p = matplotlib.patches.Circle(
                        self.pixelxy[i], radius=self.pixel_radius,
                        facecolor=cmap(norm(data[i])),
                        linewidth=2, zorder=0)
                    self.ax.add_patch(p)
                    self.patches.append(p)
                    self.select_pixel(i, self.roi[i])
                m = 8 * self.pixel_radius
                self.ax.set_xlim(minxy[0] - m, maxxy[0] + m)
                self.ax.set_ylim(maxxy[1] + m, minxy[1] - m)
            else:
                for i in range(self.pixels):
                    self.patches[i].set_facecolor(cmap(norm(data[i])))
        else:
            data2d = data.reshape((self.wh[1], self.wh[0]))
            if self.data is None:
                im = self.ax.imshow(data2d, cmap=self.cmap, zorder=-1,
                                    origin='lower')
                im.set_extent((0, self.wh[0], 0, self.wh[1]))
                # Turn roi into rgba
                roi2d = np.outer(self.roi, self.roi_alpha).reshape(
                    self.wh[1], self.wh[0], 4)
                rim = self.ax.imshow(roi2d, zorder=0, origin='lower')
                rim.set_extent((0, self.wh[0], 0, self.wh[1]))
            else:
                im = self.ax.get_images()[0]
                im.set_data(data2d)
                # im.set_clim(data.min(), data.max())
                im.set_extent((0, self.wh[0], 0, self.wh[1]))
                im.autoscale()
            self.ax.autoscale()
            self.pixel_radius = 1.5
            m = 1.5 * self.pixel_radius
            self.ax.set_xlim(-m, self.wh[0] + m)
            self.ax.set_ylim(-m, self.wh[1] + m)

        self.data = data
        self.draw_idle()

    @pyqtSlot(str)
    def set_cmap(self, s):
        "Set colors for heatmaps"
        self.cmap = s
        for im in self.ax.get_images():
            im.set_cmap(s)
        self.draw_idle()

    def stop_polygon(self):
        "Remove polygon lines"
        if self.polyline is not None:
            self.ax.lines.remove(self.polyline)
            self.polyline = None
        if self.polystart is not None:
            self.polystart.remove()
            self.polystart = None

    def select_pixel(self, i, sign):
        self.roi[i] = sign
        self.patches[i].set_edgecolor('red' if sign else None)

    def update_roi_image(self):
        if self.data is None:
            return
        if self.pixelxy is not None:
            for i in range(self.pixels):
                self.select_pixel(i, self.roi[i])
        else:
            roi2d = np.outer(self.roi, self.roi_alpha).reshape(
                self.wh[1], self.wh[0], 4)
            self.ax.get_images()[1].set_data(roi2d)

    def invert_roi(self):
        if self.data is None:
            return
        self.roi = ~self.roi
        self.update_roi_image()
        self.emit_updated()
        self.draw_idle()

    def clear(self):
        if self.data is None:
            return
        if self.pixelxy is not None:
            for i in range(self.pixels):
                if self.roi[i]:
                    self.select_pixel(i, False)
        else:
            self.roi.fill(False)
            self.update_roi_image()
        self.stop_polygon()
        self.emit_updated()
        self.draw_idle()

    def get_draw_mode(self):
        return self.drawmode

    def set_draw_mode(self, mode):
        assert mode in ['click', 'add', 'remove']
        if mode == 'click' and self.drawmode != mode:
            self.stop_polygon()
            self.draw_idle()
        self.drawmode = mode

    def erase_last_point(self):
        "Remove the last point from the unfinished polygon"
        if self.polyline is not None:
            xy = self.polyline.get_data(orig=True)
            if len(xy[0]) > 1:
                self.polyline.set_data(xy[0][:-1], xy[1][:-1])
            else:
                self.stop_polygon()
            self.draw_idle()

    def polygon_click(self, event):
        "Add polygon point"
        if event.button == 3: # Right
            self.erase_last_point()
        if event.button != 1: # Left
            self.drawing = False
            return

        upd = False
        exy = event.xdata, event.ydata
        if self.polyline is None:
            self.polyline = self.ax.plot(
                [exy[0]], [exy[1]], lw=1, color='k', zorder=1)[0]
            self.polystart = matplotlib.patches.RegularPolygon(
                exy, numVertices=5, radius=self.pixel_radius, lw=2,
                ec='black', fc='yellow', fill=True, zorder=2)
            self.ax.add_patch(self.polystart)
            self.drawing = True
        elif self.polystart.contains(event)[0]:
            pxy = self.polyline.get_xydata()
            if len(pxy) > 2:
                sign = self.drawmode == 'add'
                if self.pixelxy is None:
                    ppx, ppy = skimage.draw.polygon(
                        pxy[:,0], pxy[:,1], shape=self.wh)
                    pp = ppx + ppy * self.wh[0]
                    self.roi[pp] = sign
                    self.update_roi_image()
                else:
                    pg = matplotlib.patches.Polygon(pxy)
                    for i in range(self.pixels):
                        if pg.contains_point(self.pixelxy[i]):
                            self.select_pixel(i, sign)
            # Do nothing only if we just started drawing
            if len(pxy) > 1 or not self.drawing:
                upd = True
                self.stop_polygon()
                self.drawing = False
        else:
            xy = self.polyline.get_data(orig=True)
            d2 = (xy[0][-1] - exy[0]) ** 2 + (xy[1][-1] - exy[1]) ** 2
            if not self.drawing or d2 > self.pixel_radius ** 2:
                self.polyline.set_data(np.append(xy[0], exy[0]),
                                       np.append(xy[1], exy[1]))
                self.drawing = True
        if upd:
            self.emit_updated()
        self.draw_idle()

    def regular_click(self, event):
        upd = False
        if self.pixelxy is not None:
            for i in range(self.pixels):
                if self.patches[i].contains(event)[0]:
                    if not self.drawing:
                        self.drawsign = not self.roi[i]
                        self.drawing = True
                    if self.roi[i] != self.drawsign:
                        self.select_pixel(i, self.drawsign)
                        upd = True
        else:
            x, y = (math.floor(event.xdata), math.floor(event.ydata))
            if (0 <= x < self.wh[0] and 0 <= y < self.wh[1]):
                i = x + y * self.wh[0]
                if not self.drawing:
                    self.drawsign = not self.roi[i]
                    self.drawing = True
                if self.roi[i] != self.drawsign:
                    self.roi[i] = self.drawsign
                    upd = True
                    self.update_roi_image()
        if upd:
            self.emit_updated()
            self.draw_idle()

    def on_click(self, event):
        if self.data is None or event.inaxes != self.ax:
            return
        if self.drawing: # Click while drawing? We'd better just stop.
            self.drawing = False
        elif self.drawmode == 'click':
            self.regular_click(event)
        else:
            self.polygon_click(event)

    def on_release(self, event):
        self.drawing = False

    def on_move(self, event):
        if not self.drawing or event.inaxes != self.ax:
            return
        if self.drawmode == 'click':
            self.regular_click(event)
        else:
            self.polygon_click(event)


class DecompositionPlotWidget(BasePlotWidget):
    displayModesUpdated = pyqtSignal(list)

    def __init__(self, parent=None):
        super().__init__(parent)

        self.window_title = 'Decomposition'
        # Data to plot
        self.wn = None
        self.roi = None
        # What/how to plot
        self.discrete_cmap = 'tab10'
        self.display_mode = 0 # Index for now - should perhaps be string
        self.clear_data()

    def clear_data(self):
        self.error_log = []
        self.init_concentrations = None
        self.init_spectra = None
        self.concentrations = None
        self.spectra = None
        self.cluster_maps = {}
        self.update_display_modes()

    def mousePressEvent(self, event):
        self.popOut()

    def update_display_modes(self):
        "Create/update the list of display modes and signal the change"
        m = ['Error progress']
        if self.concentrations is not None:
            m = m + [ 'Initial spectra', 'Spectra', 'Contributions' ]
            for i in range(len(self.concentrations)):
                m.append('Contribution %d' % (i+1))
            m = m + list(self.cluster_maps.keys())
        self.display_modes = m
        if self.display_mode <= len(m):
            self.display_mode = len(m) - 1
        self.displayModesUpdated.emit(self.display_modes)

    @pyqtSlot(int)
    def set_display_mode(self, m):
        assert m < len(self.display_modes)
        self.display_mode = m
        self.draw_idle()

    def set_basic_data(self, wn, wh, pixelxy):
        "Set wavenumbers and geometric data that won't change for this file"
        if wh is None and pixelxy is None:
            raise ValueError('Either wh or pixelxy must be specified')
        if pixelxy is not None:
            pixels = len(pixelxy)
        else:
            pixels = wh[0] * wh[1]
        self.wh = wh
        self.pixelxy = pixelxy
        self.pixels = pixels
        self.wn = wn
        self.clear_and_set_roi()

    def adjust_geometry(self, wh):
        "Change shape without altering pixel count"
        pixels = wh[0] * wh[1]
        assert pixels == self.pixels
        self.wh = wh
        self.draw_idle()

    def clear_and_set_roi(self, roi=None):
        "Set ROI and clear spectra/concentrations"
        if roi is not None:
            assert len(roi) == self.pixels
        self.roi = roi
        self.clear_data()
        self.draw_idle()

    def set_concentrations(self, concentrations):
        """
        Set concentration matrix from decomposition

        Parameters
        ----------
        concentrations : array(ncomponents, npixels)
            Note the order.

        Returns
        -------
        None.
        """
        assert self.wn is not None
        if concentrations is None:
            self.concentrations = None
            self.draw_idle()
            return
        comps, pixels = concentrations.shape
        udm = self.concentrations is None or len(self.concentrations) != comps
        if self.roi is None:
            assert pixels == self.pixels
            self.concentrations = concentrations
        else:
            assert pixels == self.roi.sum()
            if udm:
                self.concentrations = np.empty((comps, self.pixels))
            self.concentrations[...] = concentrations.min(1)[:,None]
            self.concentrations[:, self.roi] = concentrations

        self.add_clustering('Strongest contribution',
                            np.argmax(self.concentrations, axis=0))
        if udm:
            self.update_display_modes()
        self.draw_idle()

    def get_concentrations(self):
        return self.concentrations

    def set_spectra(self, spectra):
        self.spectra = spectra
        self.draw_idle()

    def set_initial_concentrations(self, concentrations):
        self.init_concentrations = concentrations
        self.draw_idle()

    def set_initial_spectra(self, spectra):
        self.init_spectra = spectra
        self.draw_idle()

    def set_errors(self, errors):
        "Set the list/array of errors"
        self.error_log = errors
        if self.display_mode == 0:
            self.draw_idle()

    @pyqtSlot('QString')
    def set_cmap(self, s):
        "Set colors for heatmaps"
        self.cmap = s
        self.draw_idle()

    @pyqtSlot('QString')
    def set_discrete_cmap(self, s):
        "Set colors for clustering etc"
        self.discrete_cmap = s
        self.draw_idle()

    def add_clustering(self, name, clustered):
        assert len(clustered) == self.pixels
        self.cluster_maps[name] = clustered
        if name not in self.display_modes:
            self.update_display_modes()

    def draw_error_log(self, ax):
        ax.set_yscale('log')
        ax.xaxis.set_major_locator(
            matplotlib.ticker.MaxNLocator(integer=True))
        ax.set_xlabel('Iteration')
        if self.error_log is None:
            return
        if len(self.error_log):
            ax.plot(self.error_log, label='Error')
            ax.legend()

    def draw_init_spectra(self, ax):
        ax.set_yscale('linear')
        if self.init_spectra is not None:
            for i, s in enumerate(self.init_spectra):
                ax.plot(self.wn, s,
                        label='Initial component %d' % (i + 1))
            ax.legend()

    def draw_spectra(self, ax):
        ax.set_yscale('linear')
        if self.spectra is not None:
            for i, s in enumerate(self.spectra):
                ax.plot(self.wn, s, label='Component %d' % (i + 1))
            ax.legend()

    def draw_concentrations(self, ax):
        ax.set_yscale('linear')
        if self.concentrations is not None:
            for i, s in enumerate(self.concentrations):
                if s.any():
                    ax.plot(s, label='Component %d' % (i + 1))
            ax.legend()

    def draw_heatmap(self, ax, data, discrete=False, cbfig=None):
        if discrete:
            lc = int(data.max())
            cminmax = (0, lc)
            cmap = self.discrete_cmap
            cm = plt.get_cmap(cmap)
            if hasattr(cm, 'colors') and cm.N < 32:
                # Reshape color list to match the input values
                cmap = matplotlib.colors.ListedColormap(cm.colors, N=lc+1)
        else:
            cmap = plt.get_cmap(self.cmap)
            cminmax = data.min(), data.max()

        if self.pixelxy is not None:
            norm = matplotlib.colors.Normalize(
                vmin=cminmax[0], vmax=cminmax[1])
            patches = ax.patches
            if patches:
                for i in range(len(patches)):
                    patches[i].set_facecolor(cmap(norm(data[i])))
            else:
                minxy = np.min(self.pixelxy, axis=0)
                maxxy = np.max(self.pixelxy, axis=0)
                r = 3
                ax.clear()
                ax.set_aspect('equal')
                for i in range(len(self.pixelxy)):
                    xy = self.pixelxy[i]
                    p = matplotlib.patches.Circle(
                        xy, radius=r, facecolor=cmap(norm(data[i])))
                    ax.add_patch(p)
                m = 3 * r
                ax.set_xlim(minxy[0] - m, maxxy[0] + m)
                ax.set_ylim(minxy[1] - m, maxxy[1] + m)
                # ax.set_extent((minxy[1], maxxy[1], maxxy[0], minxy[0]))
        else:
            data = data.reshape((self.wh[1], self.wh[0]))
            imgs = ax.get_images()
            if imgs:
                im = imgs[0]
                im.set_data(data)
                im.set_cmap(cmap)
                im.set_clim(*cminmax)
                im.set_extent((-.5, self.wh[0]-.5, -.5, self.wh[1]-.5))
                im.autoscale()
            else:
                ax.clear()
                img = ax.imshow(data, cmap=cmap, zorder=0, aspect='equal',
                          vmin=cminmax[0], vmax=cminmax[1], origin='lower')
                if cbfig is not None:
                    cbfig.colorbar(img, ax=ax)

    def draw(self):
        self.draw_ax(self.ax)
        FigureCanvas.draw(self)

    def popOut(self):
        # obj_name = '%s%d' % (self.objectName(), self.display_mode)
        obj_name = None # New figure on every click
        fig = plt.figure(obj_name, tight_layout=dict(pad=.6))
        fig.canvas.set_window_title(self.window_title + ' ' +
                                    self.display_modes[self.display_mode])
        ax = fig.gca()
        self.draw_ax(ax, fig=fig)
        fig.canvas.draw_idle()
        fig.show()

    def draw_ax(self, ax, fig=None):
        mode = self.display_mode
        assert mode < len(self.display_modes)
        plotmodes = 4
        if mode < plotmodes:
            ax.clear()
            ax.set_aspect('auto')
            if mode == 0:
                self.draw_error_log(ax)
            elif mode == 1:
                self.draw_init_spectra(ax)
            elif mode == 2:
                self.draw_spectra(ax)
            elif mode == 3:
                self.draw_concentrations(ax)
        else:
            cnum = mode - plotmodes
            if cnum < len(self.concentrations):
                self.draw_heatmap(ax, self.concentrations[cnum],
                                  discrete=False, cbfig=fig)
            else:
                data = list(self.cluster_maps.values())[
                    cnum - len(self.concentrations)]
                self.draw_heatmap(ax, data, discrete=True, cbfig=fig)


class ClusterPlotWidget(BasePlotWidget):
    clicked = pyqtSignal(int)  # Clicked on cluster N

    def __init__(self, parent=None):
        super().__init__(parent)

        self.window_title = 'Clustering'
        # Data to plot
        self.clusters = None
        self.roi = None
        self.cmap = 'tab10'
        self.mpl_connect('button_press_event', self.on_click)

        # Things common to both display modes
        self.ax.set_aspect('equal')
        self.ax.tick_params(bottom=False, labelbottom=False,
                            left=False, labelleft=False)
        self.ax.set_facecolor('#eeeeee')


    def set_basic_data(self, wh=None, pixelxy=None):
        "Set shape/points and clears clusters"
        if wh is None and pixelxy is None:
            raise ValueError('Either wh or pixelxy must be specified')
        if pixelxy is not None:
            pixels = len(pixelxy)
        else:
            pixels = wh[0] * wh[1]
        self.wh = wh
        self.pixelxy = pixelxy
        self.pixels = pixels
        # clusters_of_all: 0 or cluster+1 for all pixels
        self.clusters_of_all = np.zeros((self.pixels), dtype=np.uint8)
        self.set_roi_and_clusters(None, None)

    def adjust_geometry(self, wh):
        "Update the plot geometry, keeping the pixel count"
        pixels = wh[0] * wh[1]
        assert pixels == self.pixels
        self.wh = wh
        self.update_image()

    @pyqtSlot('QString')
    def set_cmap(self, s):
        "Set colors for clustering etc"
        self.cmap = s
        self.draw_idle()

    def clear_clusters(self):
        self.set_roi_and_clusters(None, None)

    def cluster_color(self, cluster):
        return self.cmap_object(self.norm(cluster))

    @pyqtSlot(np.ndarray)
    def set_roi_and_clusters(self, roi, clusters):
        "Set ROI and cluster labels for all pixels in ROI"
        if roi is not None:
            if len(roi) != self.pixels:
                raise ValueError('ROI size must match the current geometry')
        setup_axes = self.clusters is None
        self.roi = roi
        self.clusters = clusters
        if clusters is None:
            self.clusters_of_all.fill(0)
            self.ax.clear()
            self.draw_idle()
            return

        if roi is None:
            self.clusters_of_all[:] = clusters + 1
        else:
            self.clusters_of_all.fill(0)
            self.clusters_of_all[roi] = clusters + 1

        if setup_axes:
            ...

        lc = clusters.max()
        cminmax = (0, lc)
        cmap = plt.get_cmap(self.cmap)
        if hasattr(cmap, 'colors') and cmap.N < 32:
            # Reshape color list to match the input values
            cmap = matplotlib.colors.ListedColormap(cmap.colors, N=lc+1)
        self.cmap_object = cmap
        self.norm = matplotlib.colors.Normalize(
            vmin=cminmax[0], vmax=cminmax[1])

        if self.pixelxy is not None:
            unroi = (.5, .5, .5)
            if setup_axes:
                minxy = np.min(self.pixelxy, axis=0)
                maxxy = np.max(self.pixelxy, axis=0)
                r = (maxxy - minxy).max() * .02
                for i, xy in enumerate(self.pixelxy):
                    c = self.clusters_of_all[i]
                    fc = self.cluster_color(c - 1) if c else unroi
                    p = matplotlib.patches.Circle(xy, radius=r, facecolor=fc)
                    self.ax.add_patch(p)
                m = 3 * r
                self.ax.set_xlim(minxy[0] - m, maxxy[0] + m)
                self.ax.set_ylim(minxy[1] - m, maxxy[1] + m)
            else:
                for i, p in enumerate(self.ax.patches):
                    c = self.clusters_of_all[i]
                    p.set_facecolor(self.cluster_color(c - 1) if c else unroi)
        else:
            i2d = np.zeros((self.wh[1] * self.wh[0], 4))
            i2d[roi, :] = self.cluster_color(clusters)
            i2d = i2d.reshape((self.wh[1], self.wh[0], 4))
            if setup_axes:
                self.ax.imshow(i2d, zorder=0, origin='lower')
            else:
                im = self.ax.get_images()[0]
                im.set_data(i2d)
        self.draw_idle()

    def on_click(self, event):
        if self.clusters is None or event.inaxes != self.ax:
            return
        if self.pixelxy is not None:
            for i, p in enumerate(self.ax.patches):
                if self.roi is None or self.roi[i]:
                    if p.contains(event)[0]:
                        self.clicked.emit(self.clusters_of_all[i] - 1)
        else:
            x, y = (int(event.xdata + .5), int(event.ydata + .5))
            if (0 <= x < self.wh[0] and 0 <= y < self.wh[1]):
                i = x + y * self.wh[0]
                c = self.clusters_of_all[i]
                if c:
                    self.clicked.emit(c - 1)




"""
Built-in least squares / regression methods.

All models will follow the formalism, AX = B, solve for X.

NOTE: coef_ will be X.T, which is the formalism that scikit-learn follows

"""

from abc import (ABC as _ABC, abstractmethod as _abstractmethod)

import numpy as _np

from scipy.linalg import lstsq as _lstsq
from scipy.optimize import nnls as _nnls

class LinearRegression(_ABC):
    """ Abstract class for linear regression methods """
    def __init__(self):
        self.X_ = None
        self.residual_ = None

    @property
    def coef_(self):
        """ The transposed form of X. This is the formalism of scikit-learn """
        if self.X_ is None:
            return None
        else:
            return self.X_.T

    @_abstractmethod
    def fit(self, A, B):
        """ AX = B, solve for X """

class OLS(LinearRegression):
    """
    Ordinary least squares regression

    AX = B, solve for X (coefficients.T)

    Attributes
    ----------
    coef_ : ndarray
        Regression coefficients (X.T)

    residual_ : ndarray
        Residual (sum-of-squares)

    rank_ : int
        Effective rank of matrix A

    svs_ : ndarray
        Singular values of matrix A

    Notes
    -----
    This is simply a wrapped version of Ordinary Least Squares
    (scipy.linalg.lstsq).

    coef_ is X.T, which is the formalism of scikit-learn

    """
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.rank_ = None
        self.svs_ = None

    def fit(self, A, B):
        """ Solve for X: AX = B"""
        self.X_, self.residual_, self.rank_, self.svs_ = _lstsq(A, B)

class NNLS(LinearRegression):
    """
    Non-negative constrained least squares regression

    AX = B, solve for X (coeffients.T)

    Attributes
    ----------
    coef_ : ndarray
        Regression coefficients

    residual_ : ndarray
        Residual (sum-of-squares)

    Notes
    -----
    This is simply a wrapped version of NNLS
    (scipy.optimize.nnls).

    coef_ is X.T, which is the formalism of scikit-learn
    """
    def __init__(self, *args, **kwargs):
        super().__init__()

    def fit(self, A, B):
        """ Solve for X: AX = B"""

        if B.ndim == 2:
            N = B.shape[-1]
        else:
            N = 0

        self.X_ = _np.zeros((A.shape[-1], N))
        self.residual_ = _np.zeros((N))

        # nnls is Ax = b; thus, need to iterate along
        # columns of B
        if N == 0:
            self.X_, self.residual_ = _nnls(A, B)
        else:
            for num in range(N):
                self.X_[:, num], self.residual_[num] = _nnls(A, B[:, num])



import setuptools
import os.path
from setuptools.command.install import install

executables = {'preprocessing': 'Preprocessing',
               'mcr_als': 'MCR-ALS',
               'clustering': 'Clustering'}

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="octavvs",
    version="0.1.10",
    author="Syahril Siregar, Carl Troein, Michiel Op De Beeck et al.",
    author_email="carl@thep.lu.se",
    description="Open Chemometrics Toolkit for Analysis and Visualization of Vibrational Spectroscopy data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ctroein/octavvs",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['numpy', 'scipy', 'matplotlib', 'sklearn', 'pyshortcuts',
                      'opencv-python', 'pandas', 'pillow', 'pymatreader', 'dill',
                      'threadpoolctl'],
    package_data={ '': ['*.ui', '*.mat', '*.ico', '*.icns'] },
    entry_points={'console_scripts':
        ['octavvs = octavvs.launcher:main',
        'oct_preprocessing = octavvs.preprocessing:main',
        'oct_decomposition = octavvs.decomposition:main',
        'oct_make_icons = octavvs.make_icons:main'],
    },

)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  2 23:00:44 2021

@author: carl
"""

import os.path
import numpy as np
import h5py
import csv
import traceback
import time
import scipy.io

from .spectraldata import SpectralData

class DecompositionData(SpectralData):
    "SpectralData with addition of ROI/decomposition data"
    def __init__(self):
        super().__init__()
        # self.rdc_directory = None # None means dirname(curFile)
        self.clear_rdc()

    def clear_rdc(self):
        """
        Clear the ROI/decomposition/clustering data.

        Returns
        -------
        None.
        """
        self.roi = None # bool array or None if nothing is selected
        self.decomposition_settings = None # dict with decomp settings
        self.decomposition_roi = None # ROI actually used for decomp, or None
        # {iteration: {'spectra': array(component, wavenum),
        # 'concentrations': array(component, pixel) } }
        self.decomposition_data = None
        self.decomposition_errors = None # 1d array of error values
        self.clustering_settings = None # dict with clustering settings
        self.clustering_roi = None # ROI actually used for clustering, or None
        self.clustering_labels = None # 1d int array of labels
        self.clustering_annotations = None # {str: [labels]}

    # def set_rdc_directory(self, directory):
    #     self.rdc_directory = directory

    def get_duplicate_filenames(self):
        """
        Get list of filenames whose basename is not unique.

        Returns
        -------
        List of filenames with full paths
        """
        dups = set()
        seen = {}
        for i, f in enumerate([os.path.splitext(os.path.basename(f))[0]
                               for f in self.filenames]):
            if f in seen:
                dups.add(self.filenames[i])
                dups.add(self.filenames[seen[f]])
            else:
                seen[f] = i
        return list(dups)

    # def rdc_filename(self, filetype='odd'):
    #     """
    #     Get the default name of the file where the current file's
    #     ROI, decomposition and clustering is stored.

    #     Parameters
    #     ----------
    #     filedir : str, optional
    #         The directory the file will be in, or None to use the input dir.

    #     Returns
    #     -------
    #     str
    #         Filename.
    #     """
    #     filetypes = ['odd']
    #     if filetype not in filetypes:
    #         raise ValueError("File type must be one of " + str(filetypes))
    #     if self.rdc_directory is None:
    #         return os.path.splitext(self.curFile)[0] + '.' + filetype
    #     if self.rdc_directory == '':
    #         raise ValueError("Directory cannot be ''")
    #     fn = os.path.splitext(os.path.basename(self.curFile))[0]
    #     return os.path.join(self.rdc_directory, fn) + '.' + filetype

    def read_matrix(self, filename):
        self.clear_rdc()
        super().read_matrix(filename)
        # try:
        #     self.load_rdc()
        # except OSError:
        #     pass
        # except Exception:
        #     traceback.print_exc()
        #     print('Warning: rdc auto-load failed')


    # ROI stuff
    def set_roi(self, roi):
        "Set ROI from array."
        if roi is None:
            self.roi = None
        elif roi.shape != (self.raw.shape[0],):
            raise ValueError('Wrong number of values for ROI (%s)' %
                             str(roi.shape))
        else:
            self.roi = roi if roi.any() else None

    # Decomposition stuff
    def set_decomposition_settings(self, params):
        """
        Populate decomposition_settings from a Parameters object,
        set the decomposition ROI depending on params.dcRoi,
        and clear stored data.

        Parameters
        ----------
        params : Parameters
            Only the dc settings will be copied, without the 'dc' prefix.

        Returns
        -------
        None

        Raises
        -------
        ValueError
            If the dcRoi parameter is inconsistent with current ROI
        """
        dcs = params.filtered('dc')
        roi = None
        if params.dcRoi == 'ignore':
            roi = None
        elif params.dcRoi == 'require':
            if self.roi is None:
                raise ValueError('ROI is required but not defined')
            roi = self.roi.copy()
        elif params.dcRoi == 'ifdef':
            roi = self.roi.copy() if self.roi is not None else None
        else:
            raise ValueError('Bad value for params.dcRoi')

        self.decomposition_settings = {k[2:]: v for k, v in dcs.items()}
        self.decomposition_settings['DateTime'] = \
            time.strftime('%Y-%m-%d %H:%M:%S %Z')
        self.decomposition_roi = roi
        self.decomposition_data = {}

    def set_decomposition_errors(self, errors):
        "Set the history of least squares error values"
        self.decomposition_errors = np.asarray(errors)

    def add_decomposition_data(self, iteration, spectra, concentrations):
        "Set data for a specific iteration; 0=initial"
        pixels = self.decomposition_roi.sum() if self.decomposition_roi \
            is not None else self.raw.shape[0]
        comps = self.decomposition_settings['Components']
        assert iteration <= self.decomposition_settings['Iterations']
        if iteration:
            assert spectra is not None and concentrations is not None
        else:
            assert spectra is not None or concentrations is not None
        if spectra is not None:
            assert spectra.shape == (comps, self.raw.shape[1])
            spectra = spectra.copy()
        if concentrations is not None:
            assert concentrations.shape == (comps, pixels)
            concentrations = concentrations.copy()
        self.decomposition_data[iteration] = {
            'spectra': spectra, 'concentrations': concentrations}

    # Clustering/annotation stuff
    def set_clustering_settings(self, params, roi):
        """
        Populate decomposition_settings from a Parameters object,
        copy the decomposition ROI, and clear stored clustering data.

        Parameters
        ----------
        params : Parameters
            Only the ca settings will be copied, without the 'ca' prefix.

        Returns
        -------
        None
        """
        cas = params.filtered('ca')
        self.clustering_settings = {k[2:]: v for k, v in cas.items()}
        self.clustering_settings['DateTime'] = \
            time.strftime('%Y-%m-%d %H:%M:%S %Z')
        self.clustering_roi = roi.copy() if roi is not None else None
        self.clustering_labels = None
        self.clustering_label_set = None
        self.clustering_annotations = {}

    def set_clustering_labels(self, labels, relabel):
        pixels = self.clustering_roi.sum() if \
            self.clustering_roi is not None else self.raw.shape[0]
        assert labels.shape == (pixels,)
        # Count labels, sort indexes and invert permutation array
        bc = np.bincount(labels)
        mapp = np.empty(bc.size, bc.dtype)
        mapp[np.argsort(bc)[::-1]] = np.arange(bc.size)
        self.clustering_labels = mapp[labels]
        self.clustering_label_set = set(self.clustering_labels)

    def get_annotation_clusters(self, annotation):
        if self.clustering_annotations is None:
            self.clustering_annotations = {}
        if annotation not in self.clustering_annotations:
            self.clustering_annotations[annotation] = set()
        return self.clustering_annotations[annotation].copy()

    def get_unannotated_clusters(self):
        used = set()
        for c in self.clustering_annotations.values():
            used.update(c)
        return self.clustering_label_set - used

    def set_annotation_clusters(self, annotation, clusters=None):
        self.get_annotation_clusters(annotation)
        if clusters is not None:
            cset = set(clusters)
            self.clustering_annotations[annotation] = cset
            for c in self.clustering_annotations.values():
                if c is not cset and cset.intersection(c):
                    print('Error: duplicated cluster id',
                          self.clustering_annotations)

    def del_annotation(self, annotation):
        if self.clustering_annotations is not None and \
                annotation in self.clustering_annotations:
                del self.clustering_annotations[annotation]

    # Loading and saving
    def load_rdc_(self, f, what, imgnum):
        grp = f['SpectralData/Image_%d' % imgnum]

        if what in ['all', 'roi']:
            if 'ROI' in grp:
                roi = grp['ROI']
                if roi.shape != (self.raw.shape[0],):
                    raise ValueError('Wrong number of values for ROI (%d)' %
                                     (len(roi)))
                roi = roi[:]
            else:
                roi = None

        if what in ['all', 'decomposition']:
            dcroi = None
            dcerrors = None
            if 'Decomposition' in grp:
                dc = grp['Decomposition']
                dcsettings = dict(dc.attrs.items())
                pixels = 0
                if 'ROI' in dc:
                    dcroi = dc['ROI'][:]
                    if dcroi.shape != (self.raw.shape[0],):
                        raise ValueError(
                            'Wrong number of values for '
                            'Decomposition ROI (%d)' % (len(dcroi)))
                    pixels = dcroi.sum()
                if not pixels:
                    pixels = self.raw.shape[0]
                comps = dcsettings['Components']
                dcdata = {}
                for ds in dc['Data'].values():
                    it = ds.attrs['Iteration']
                    conc = None
                    if 'Concentrations' in ds:
                        conc = ds['Concentrations'][:]
                        if conc.shape[1] != pixels:
                            raise ValueError('Concentration pixel count mismatch')
                        if conc.shape[0] != comps:
                            raise ValueError(
                                'Concentration component count mismatch')
                    spect = None
                    if 'Spectra' in ds:
                        spect = ds['Spectra'][:]
                        if spect.shape[1] != self.raw.shape[1]:
                            raise ValueError('Spectra wavenumber count mismatch')
                        if spect.shape[0] != comps:
                            raise ValueError('Spectra component count mismatch')
                    dcdata[it] = {'concentrations': conc, 'spectra': spect}
                if 'Errors' in dc:
                    dcerrors = dc['Errors'][:]
                    # if len(dcerrors) >= dcsettings['Iterations']:
                    #     raise ValueError('Too many elements in Errors')
            else:
                dcsettings = None
                dcdata = None

        if what in ['all', 'clustering']:
            caroi = None
            calabels = None
            casettings = None
            caannot = None
            if 'Clustering' in grp:
                ca = grp['Clustering']
                casettings = dict(ca.attrs.items())
                pixels = 0
                if 'ROI' in ca:
                    caroi = ca['ROI'][:]
                    if caroi.shape != (self.raw.shape[0],):
                        raise ValueError(
                            'Wrong number of values for '
                            'Clustering ROI (%d)' % (len(caroi)))
                    pixels = caroi.sum()
                if not pixels:
                    pixels = self.raw.shape[0]
                maxclust = casettings['Clusters']
                if 'Labels' in ca:
                    calabels = ca['Labels'][:]
                    if calabels.shape != (pixels,):
                        raise ValueError('Cluster label count mismatch')
                    if calabels.max() < 0 or calabels.max() >= maxclust:
                        raise ValueError('Cluster label range error')
                if 'Annotations' in ca:
                    caannot = {}
                    for ann in ca['Annotations'].values():
                        atxt = ann.attrs['Text']
                        avals = ann[:]
                        if len(avals) > 0 and (
                                min(avals) < 0 or np.max(avals) >= maxclust):
                            raise ValueError('Annotation cluster range error')
                        if atxt in caannot:
                            caannot[atxt].update(set(avals))
                        else:
                            caannot[atxt] = set(avals)

        if what in ['all', 'roi']:
            self.roi = roi
        if what in ['all', 'decomposition']:
            self.decomposition_settings = dcsettings
            self.decomposition_roi = dcroi
            self.decomposition_data = dcdata
            self.decomposition_errors = dcerrors
        if what in ['all', 'clustering']:
            self.clustering_settings = casettings
            self.clustering_roi = caroi
            self.clustering_labels = calabels
            self.clustering_label_set = set(calabels) if \
                calabels is not None else None
            # if caannot is not None:
            #     for k, v in caannot.items():
            #         caannot[k] = list(v)
            self.clustering_annotations = caannot

    def load_rdc(self, filename, *, what='all'):
        """
        Load ROI/decomp/clust from a named file.

        Parameters
        ----------
        filename : str
            A named file
        what : str
            One of 'all', 'roi', 'decomposition', 'clustering'

        Returns
        -------
        success : bool
            Whether data was loaded

        """
        # if filename is None:
        #     filename = self.rdc_filename()
        with h5py.File(filename, mode='r') as f:
            return self.load_rdc_(f, what, 0)

    def save_rdc_(self, f, what, imgnum):
        grp = f.require_group('SpectralData')
        grp.attrs.create('Creator', 'OCTAVVS')
        grp = grp.require_group('Image_%d' % imgnum)

        def replace_data(grp, name, arr):
            if name in grp:
                if arr is not None and grp[name].shape == arr.shape:
                    grp[name][...] = arr
                    return
                del grp[name]
            if arr is not None:
                grp.create_dataset(name, data=arr)

        if what in ['all', 'roi']:
            replace_data(grp, 'ROI', self.roi)

        if what in ['all', 'decomposition'] and \
            self.decomposition_settings is not None:
            dc = grp.require_group('Decomposition')
            for k, v in self.decomposition_settings.items():
                dc.attrs[k] = v
            replace_data(dc, 'ROI', self.decomposition_roi)
            ddg = dc.require_group('Data')
            for k, ds in ddg.items():
                it = ds.attrs['Iteration']
                if it not in self.decomposition_data \
                    or k != 'Iter_%d' % it:
                    del ddg[k]
            for it, vals in self.decomposition_data.items():
                ds = ddg.require_group('Iter_%d' % it)
                ds.attrs['Iteration'] = it
                replace_data(ds, 'Concentrations', vals['concentrations'])
                replace_data(ds, 'Spectra', vals['spectra'])
            replace_data(dc, 'Errors', self.decomposition_errors)

        if what in ['all', 'clustering'] and\
            self.clustering_settings is not None:
            ca = grp.require_group('Clustering')
            for k, v in self.clustering_settings.items():
                ca.attrs[k] = v
            replace_data(ca, 'ROI', self.clustering_roi)
            replace_data(ca, 'Labels', self.clustering_labels)
            if 'Annotations' in ca:
                del ca['Annotations']
            if self.clustering_annotations is not None:
                ag = ca.require_group('Annotations')
                for i, (ann, labels) in enumerate(
                        self.clustering_annotations.items()):
                    ds = ag.create_dataset('Annotation_%d' % i,
                                           data=list(labels))
                    ds.attrs['Text'] = ann


    def save_rdc(self, *, filename, what='all'):
        """
        Save data to a named file.

        Parameters
        ----------
        filename : str
            A named file; possibly from rdc_filename
        what : str
            One of 'all', 'roi', 'decomposition', 'clustering'

        Returns
        -------
        None.

        """
        with h5py.File(filename, mode='a') as f:
            return self.save_rdc_(f, what, 0)

    def save_roi(self, *, filename, fmt='.mat'):
        """
        Save ROI mask to a named file.

        Parameters
        ----------
        filename : str
            A named file
        fmt : str
            One of '.rdc', '.roi', '.mat', '.csv'

        Returns
        -------
        None.

        """
        print(f'save "{fmt}"')
        if fmt == '.rdc' or fmt == '.roi':
            self.save_rdc(filename=filename, what='roi')
            return
        if self.roi is None:
            roi = np.zeros(self.wh, dtype=int)
        else:
            roi = self.roi.reshape(self.wh).astype(int)
        if fmt == '.mat':
            scipy.io.savemat(filename, mdict={'ROI': roi},
                             appendmat=False, do_compression=True)
        elif fmt == '.csv':
            with open(filename, 'w') as f:
                writer = csv.writer(f)
                writer.writerows(roi)
        else:
            raise ValueError('Unknown ROI save format %s' % fmt)

    def get_headers_as_lists(self):
        headers = []
        headers.append(['#Input file', self.curFile])
        if self.clustering_settings['Input'] == 'decomposition':
            for k, v in self.decomposition_settings.items():
                headers.append(['#Decomposition ' + k, v])
        for k, v in self.clustering_settings.items():
            headers.append(['#Clustering ' + k, v])
        return headers

    def save_spectra_csv(self, writer, average):
        writer.writerows(self.get_headers_as_lists())
        if average:
            unused = self.get_unannotated_clusters()
            # # If all are unannotated, output cluster labels instead
            # by_cluster = len(unused) == len(self.clustering_label_set)
            header = ['Wavenumber/%s' % average]
            avg_indexes = []
            if self.clustering_roi is None:
                roi_indexes = np.arange(len(self.raw))
            else:
                roi_indexes = self.clustering_roi.nonzero()[0]
            if average == 'cluster':
                for c in self.clustering_label_set:
                    header.append('Cluster %d' % c)
                    avg_indexes.append(
                        roi_indexes[self.clustering_labels == c])
            else:
                for a, cs in self.clustering_annotations.items():
                    if cs:
                        header.append(a)
                        avg_indexes.append(roi_indexes[
                            np.isin(self.clustering_labels, list(cs))])
                if unused:
                    header.append('(unannotated)')
                    avg_indexes.append(roi_indexes[
                        np.isin(self.clustering_labels, list(unused))])
            writer.writerow(['Nspectra'] + [len(ixs) for ixs in avg_indexes])
            writer.writerow(header)
            for i, wn in enumerate(self.wavenumber):
                row = [wn]
                for aix in avg_indexes:
                    row.append(self.raw[aix, i].mean())
                writer.writerow(row)
        else:
            cluster_annots = {c: '(unannotated)' for c in
                              self.get_unannotated_clusters()}
            for a, cc in self.clustering_annotations.items():
                for c in cc:
                    cluster_annots[c] = a
            headers = [['x'], ['y'], ['Cluster'], ['Wavenumber/Annotation']]
            if self.clustering_roi is None:
                roi_indexes = np.arange(len(self.raw))
            else:
                roi_indexes = self.clustering_roi.nonzero()[0]
            for i, ix in enumerate(roi_indexes):
                if self.pixelxy is None:
                    xy = (ix % self.wh[0], ix // self.wh[0])
                else:
                    xy = self.pixelxy[ix]
                headers[0].append(xy[0])
                headers[1].append(xy[1])
                c = self.clustering_labels[i]
                headers[2].append(c)
                headers[3].append(cluster_annots[c])
            writer.writerows(headers)
            writer.writerows(np.hstack((self.wavenumber[:,None],
                                        self.raw[roi_indexes,:].T)))

    def save_annotated_spectra(self, filename, filetype, average):
        if average not in ['annotation', 'cluster', None]:
            raise ValueError("Invalid value for 'average'")
        if filetype == 'csv':
            with open(filename, 'w', newline='') as f:
                self.save_spectra_csv(csv.writer(f), average=average)
        else:
            raise ValueError('Invalid filetype %d' % filetype)





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Feb  4 11:48:28 2019

A set of baseline correction algorithms

@author: Carl Troein
"""

import numpy as np
from scipy import sparse
from scipy.spatial import ConvexHull
from scipy.interpolate import interp1d
from scipy.sparse.linalg import spsolve
from multiprocessing.pool import Pool, ThreadPool
import os
import dill


def straight(x, y):
    """
    Return a straight line baseline correction.
    x: wavenumbers, sorted either way
    y: spectrum or spectra at those wavenumbers; shape (..., wavenumber)
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    # Create baseline using linear interpolation between vertices
    if x[0] < x[-1]:
        return interp1d(x[[0,-1]], y[...,[0,-1]], assume_sorted=True)(x)
    return interp1d(x[[-1,0]], y[...,[-1,0]], assume_sorted=True)(x)


def apply_packed_function_for_map(dumped):
    "Unpack dumped function as target function and call it with arguments."
    return dill.loads(dumped[0])(dumped[1])

def pack_function_for_map(target_function, items):
    dumped_function = dill.dumps(target_function)
    dumped_items = [(dumped_function, item) for item in items]
    return apply_packed_function_for_map, dumped_items


def mp_bgcorrection(func, y, lim_single=8, lim_tp=40, progressCallback=None):
    if len(y) < 1:
        return y.copy()
    if y.ndim < 2:
        return func(y)
    if hasattr(os, 'sched_getaffinity'):
        cpus = len(os.sched_getaffinity(os.getpid()))
    else:
        cpus = os.cpu_count()
    cpus = min(cpus, len(y))
    if cpus == 1 or len(y) <= lim_single:
        cpus = 1
        it = map(func, y)
    elif len(y) <= lim_tp:
        cpus = min(cpus, 3)
        pool = ThreadPool(cpus)
        it = pool.imap(func, y, chunksize=5)
    else:
        pool = Pool(cpus)
        it = pool.imap(*pack_function_for_map(func, y), chunksize=10)

    ret = np.empty_like(y)
    for i in range(len(y)):
        ret[i] = next(it)
        if progressCallback:
            progressCallback(i+1, len(y))
    return ret

def asls(y, lam, p, niter=20, progressCallback=None):
    """
    Return the baseline computed by Asymmetric least squares background correction, AsLS.
    Ref: Baseline correction with asymmetric least squares smoothing. PHC Eilers & HFM Boelens.
        Leiden University Medical Centre Report, 2005
    Parameters:
    y: one spectrum to correct, or multiple as an array of shape (spectrum, wavenumber)
    lam: lambda, the smoothness parameter
    p: p, the asymmetry parameter, typically .001 to 0.1
    niter: maximum number of iterations
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    L = y.shape[-1]
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    D = lam * D.dot(D.T)

    def asls_one(yy):
        w = np.ones(L)
        for i in range(niter):
            z = spsolve(sparse.spdiags(w, 0, L, L) + D, w * yy)
            wnew = (p - .5) * np.sign(yy - z) + .5
#            wnew = p * (yy > z) + (1-p) * (yy < z)
            if np.array_equal(wnew, w):
                break
            w = wnew
        return z

    return mp_bgcorrection(asls_one, y, progressCallback=progressCallback)

def iasls(y, lam, lam1, p, niter=30, progressCallback=None):
    """
    Return the baseline computed by Improved asymmetric least squares background correction, IAsLS.
    Ref: Baseline correction for Raman spectra using an improved asymmetric least squares method.
        Shixuan He, Wei Zhang, Lijuan Liu, Yu Huang, Jiming He, Wanyi Xie, Peng Wu and Chunlei Du.
        Anal. Methods, 2014, 6, 4402-4407. DOI: 10.1039/C4AY00068D
    In this implementation, W is not squared so p carries the same meaning as in AsLS.
    Parameters:
    y: one spectrum to correct, or multiple as an array of shape (spectrum, wavenumber)
    lam: lambda, the smoothness parameter
    lam1: lambda1, the 1st derivatives smoothness parameter
    p: p, the asymmetry parameter
    niter: maximum number of iterations
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    L = y.shape[-1]
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    D = lam * D.dot(D.T)
    D1 = sparse.csc_matrix(np.diff(np.eye(L), 1))
    D1 = lam1 * D1.dot(D1.T)

    def iasls_one(yy):
        w = np.ones(L)
        W = sparse.spdiags(w, 0, L, L)
#        W = W @ W.T
        z = spsolve(W + D, w * yy)
        w = p * (yy > z) + (1-p) * (yy < z)
        for i in range(niter):
            W = sparse.spdiags(w, 0, L, L)
#            W = W @ W.T
            z = spsolve(W + D + D1, (W + D1) * yy)
            wnew = p * (yy > z) + (1-p) * (yy < z)
            if np.array_equal(wnew, w):
                break
            w = wnew
        return z
    return mp_bgcorrection(iasls_one, y, progressCallback=progressCallback)


def arpls(y, lam, ratio=1e-6, niter=1000, progressCallback=None):
    """
    Return the baseline computed by asymmetric reweighted penalized least squares smoothing, arPLS.
    Ref: Baseline correction using asymmetrically reweighted penalized least squares smoothing
        Sung-June Baek, Aaron Park, Young-Jin Ahn and Jaebum Choo
        Analyst, 2015, 140, 250-257. DOI: 10.1039/C4AN01061B
    In this implementation, W is not squared so p carries the same meaning as in AsLS.
    Parameters:
    y: one spectrum to correct, or multiple as an array of shape (spectrum, wavenumber)
    lam: lambda, the smoothness parameter
    ratio: convergence criterion; target relative change in weights between iterations
    niter: maximum number of iterations
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    L = y.shape[-1]
    D = sparse.csc_matrix(np.diff(np.eye(L), 2))
    D = lam * D.dot(D.T)

    def arpls_one(yy):
        w = np.ones(L)
        for i in range(niter):
            W = sparse.spdiags(w, 0, L, L)
            z = sparse.linalg.spsolve(W + D, w * yy)
            d = yy - z
            dn = d[d < 0]
            s = dn.std()
            wt = 1. / (1 + np.exp(2 / s * (d - (2*s-dn.mean()))))
            if np.linalg.norm(w - wt) / np.linalg.norm(w) < ratio:
                break
            w = wt
        return z
    return mp_bgcorrection(arpls_one, y, progressCallback=progressCallback)



def rubberband(x, y, progressCallback=None):
    """
    Rubberband baseline correction of one or more spectra.
    Parameters:
    x: wavenumbers, sorted in either direction
    y: spectrum at those wavenumbers, or multiple spectra as array of shape (spectrum, wavenumber)
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    if x[0] > x[-1]:
        return rubberband(x[::-1], y[...,::-1],
                          progressCallback=progressCallback)[...,::-1]
    def rubberband_one(yy):
        # Find the convex hull
        v = ConvexHull(np.column_stack((x, yy))).vertices
        # Rotate convex hull vertices until they start from the lowest one
        v = np.roll(v, -v.argmin())
        # Leave only the ascending part
        v = v[:v.argmax()+1]
        # Create baseline using linear interpolation between vertices
        b = np.interp(x, x[v], yy[v])
        return b
    return mp_bgcorrection(rubberband_one, y, lim_single=100, lim_tp=10000,
                           progressCallback=progressCallback)

def concaverubberband(x, y, iters, progressCallback=None):
    """
    Concave rubberband baseline correction. This algorithm removes more than a
    straight line, alternating with normal rubberband to bring negative points
    up again. It does not converge nicely and will eat up all the data if run
    with many iterations.
    Parameters:
    x: wavenumbers, sorted from low to high (todo: implement high-to-low)
    y: spectrum at those wavenumbers
    iters: iterations to run; note that this algorithm doesn't converge nicely
    progressCallback(int a, int b): callback function called to indicated that the processing
        is complete to a fraction a/b.
    Returns: baseline of the spectrum, measured at the same points
    """
    def concaverubberband_one(yy):
        origyy = yy
        yy = yy - rubberband(x, yy);
        for i in range(iters):
            F = .1 * (yy.max() - yy.min())
            xmid = .5 * (x[-1] + x[0])
            d2 = .25 * (x[-1] - x[0]) ** 2
            yy += F * (x - xmid)**2 / d2
            yy -= rubberband(x, yy);
        return origyy - yy
    return mp_bgcorrection(concaverubberband_one, y, lim_single=30, lim_tp=500,
                           progressCallback=progressCallback)


